import logging
from typing import Optional

import neo4j
from neo4j import AsyncGraphDatabase
from typing_extensions import override

from ...vector_db.base import BaseVectorDB
from .agent import Neo4jAgent
from .interaction import Neo4jInteraction
from .memory import Neo4jMemory
from .organization import Neo4jOrganization
from .user import Neo4jUser


class Neo4jGraphInterface(
    Neo4jOrganization, Neo4jAgent, Neo4jUser, Neo4jInteraction, Neo4jMemory
):

    def __init__(
        self,
        uri: str,
        username: str,
        password: str,
        database: str,
        associated_vector_db: Optional[BaseVectorDB] = None,
        enable_logging: bool = False,
    ):
        """
        A unified interface for interacting with the Neo4j graph database.

        Args:
            uri (str): The URI of the Neo4j database.
            username (str): The username for authentication.
            password (str): The password for authentication.
            database (str): The name of the Neo4j database.
            associated_vector_db (Optional[BaseVectorDB]): The vector database to be associated with the graph for data consistency (e.g adding / deleting memories across both.)
            enable_logging (bool): Whether to enable console logging

        Example:
            ```python
            from memora.graph_db.neo4j import Neo4jGraphInterface
            from qdrant_client import AsyncQdrantClient
            from memora.vector_db.qdrant import QdrantDB

            neo4j_interface = Neo4jGraphInterface(
                uri="Neo4jURI",
                username="Neo4jUsername",
                password="Neo4jPassword",
                database="Neo4jDatabaseName",
                # Optional Association
                associated_vector_db=QdrantDB(async_client=AsyncQdrantClient(url="QDRANT_URL", api_key="QDRANT_API_KEY"))
            )
            ```
        """

        self.driver = AsyncGraphDatabase.driver(uri=uri, auth=(username, password))
        self.database = database
        self.associated_vector_db = associated_vector_db

        # Configure logging
        self.logger = logging.getLogger(__name__)
        if enable_logging:
            logging.basicConfig(level=logging.INFO)

    @override
    def get_associated_vector_db(self) -> Optional[BaseVectorDB]:
        """
        The vector database associated with the graph database, these is used inside the graph transactional blocks
        to ensure data consistency when handling memories across both stores (e.g., saving memories to the vector
        store and creating corresponding nodes in the graph db).
        """
        return self.associated_vector_db

    @override
    async def close(self):
        self.logger.info("Closing Neo4j driver")
        await self.driver.close()

    # Setup method
    @override
    async def setup(self, *args, **kwargs) -> None:
        """Sets up Neo4j database constraints and indices for the graph schema."""

        async def create_constraints_and_indexes(tx):
            self.logger.info("Creating constraints and indexes")
            # Organization node key
            await tx.run(
                """
                CREATE CONSTRAINT unique_org_id IF NOT EXISTS 
                FOR (o:Org) REQUIRE o.org_id IS NODE KEY
            """
            )

            # User node key
            await tx.run(
                """
                CREATE CONSTRAINT unique_org_user IF NOT EXISTS
                FOR (u:User) REQUIRE (u.org_id, u.user_id) IS NODE KEY
            """
            )

            # Agent node key
            await tx.run(
                """
                CREATE CONSTRAINT unique_org_agent IF NOT EXISTS 
                FOR (a:Agent) REQUIRE (a.org_id, a.agent_id) IS NODE KEY
            """
            )

            # Memory node key
            await tx.run(
                """
                CREATE CONSTRAINT unique_user_memory IF NOT EXISTS
                FOR (m:Memory) REQUIRE (m.org_id, m.user_id, m.memory_id) IS NODE KEY
            """
            )

            # Interaction node key
            await tx.run(
                """
                CREATE CONSTRAINT unique_user_interaction IF NOT EXISTS
                FOR (i:Interaction) REQUIRE (i.org_id, i.user_id, i.interaction_id) IS NODE KEY
            """
            )

            # Date node key
            await tx.run(
                """
                CREATE CONSTRAINT unique_user_date IF NOT EXISTS
                FOR (d:Date) REQUIRE (d.org_id, d.user_id, d.date) IS NODE KEY
            """
            )

        self.logger.info("Setting up Neo4j database constraints and indices")
        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(create_constraints_and_indexes)
        self.logger.info("Setup complete")

    # Migration method
    async def migrate_to_schema_for_memora_v0_3_x(self, *args, **kwargs) -> None:
        """
        Migrate the Neo4j graph database schema to the version that works with Memora v0.3.x

        This migration involves establishing :DATE_OBTAINED relationships between `Memory` nodes and `Date` nodes based on
        their `org_id, user_id, obtained_at` attributes and dropping the index `:Interaction (updated_at) -> interaction_updated_timestamp` (because
        Neo4j does not utilize it for index-backed sorting)
        """

        async def link_memories_to_dates(tx):
            self.logger.info(
                "Linking Memory nodes to Date nodes based on org_id, user_id, obtained_at"
            )
            await tx.run(
                """
                MATCH (memory:Memory)
                MERGE (date:Date {org_id: memory.org_id, user_id: memory.user_id, date: date(memory.obtained_at)})
                MERGE (memory)-[:DATE_OBTAINED]->(date)
                """
            )
            self.logger.info("Memory nodes successfully linked to their Date nodes")

        async def drop_index(tx):
            self.logger.info(
                "Dropping interaction_updated_timestamp_index index if it exists"
            )
            await tx.run(
                """
                DROP INDEX interaction_updated_timestamp_index IF EXISTS
                """
            )
            self.logger.info("Index dropped (if it existed)")

        self.logger.info(
            "Starting migration to graph schema that works with Memora v0.3.x"
        )
        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(link_memories_to_dates)
            await session.execute_write(drop_index)
        self.logger.info("Migration of graph schema completed")



from typing import Optional

from pydantic import BaseModel, Field


class ExtractedMemory(BaseModel):
    memory: str = Field(
        description="The memory, max 25 words, self-contained, remember use of #user_#id# or #agent_#id#."
    )
    msg_source_ids: list[int] = Field(
        description="List with id or ids indicating which message blocks this memory was extracted from."
    )


class MemoryExtractionResponse(BaseModel):
    memories_first_pass: Optional[list[ExtractedMemory]] = Field(
        default_factory=list,
        description="First pass of useful info written for memory (remember use of #user_#id# or #agent_#id#) with their source message ids.",
    )
    memories_second_pass: Optional[list[ExtractedMemory]] = Field(
        default_factory=list,
        description="Second pass containing info missed in first pass with their source message ids, if any.",
    )
    memories_third_pass: Optional[list[ExtractedMemory]] = Field(
        default_factory=list,
        description="Third pass containing info missed in first and second passes with their source message ids, if any.",
    )


class NewGleanedMemory(BaseModel):
    memory: str = Field(
        description="A new gleaned memory, max 25 words, self-contained."
    )
    source_candidate_pos_id: int = Field(
        description="The POS_ID of the candidate memory from which this new memory was gleaned."
    )


class ContraryMemory(BaseModel):
    memory: str = Field(
        description="The candidate memory that directly contradicting an existing memory, max 25 words, self-contained."
    )
    source_candidate_pos_id: int = Field(
        description="The POS_ID of the candidate memory from which this contrary memory was sourced."
    )
    contradicted_memory_id: str = Field(
        description="The ID of the existing memory that was contradicted."
    )


class MemoryComparisonResponse(BaseModel):
    new_memories: list[NewGleanedMemory] = Field(
        description="List of newly gleaned memories."
    )
    contrary_memories: list[ContraryMemory] = Field(
        description="List of candidate memories that contradict existing ones, if any."
    )



from .azure_openai_backend_llm import AzureOpenAIBackendLLM
from .groq_backend_llm import GroqBackendLLM
from .openai_backend_llm import OpenAIBackendLLM
from .together_backend_llm import TogetherBackendLLM

__all__ = [
    "AzureOpenAIBackendLLM",
    "GroqBackendLLM",
    "OpenAIBackendLLM",
    "TogetherBackendLLM",
]



import logging
import re
from datetime import datetime
from typing import Dict, List, Optional, Set, Tuple

import memora.schema.models as models
from memora.graph_db.base import BaseGraphDB
from memora.llm_backends.base import BaseBackendLLM
from memora.prompts import (
    COMPARE_EXISTING_AND_NEW_MEMORIES_INPUT_TEMPLATE,
    COMPARE_EXISTING_AND_NEW_MEMORIES_SYSTEM_PROMPT,
    EXTRACTION_MSG_BLOCK_FORMAT,
    FILTER_RETRIEVED_MEMORIES_SYSTEM_PROMPT,
    MEMORY_EXTRACTION_SYSTEM_PROMPT,
    MEMORY_EXTRACTION_UPDATE_SYSTEM_PROMPT,
    MSG_MEMORY_SEARCH_PROMPT,
    MSG_MEMORY_SEARCH_TEMPLATE,
)
from memora.schema.extraction_schema import (
    MemoryComparisonResponse,
    MemoryExtractionResponse,
)
from memora.schema.storage_schema import (
    ContraryMemoryToStore,
    MemoriesAndInteraction,
    MemoryToStore,
)
from memora.vector_db.base import BaseVectorDB, MemorySearchScope


class Memora:
    """
    This class orchestrates all necessary memory operations and offers a unified interface for utilizing the Memory Agent.
    """

    def __init__(
        self,
        vector_db: BaseVectorDB,
        graph_db: BaseGraphDB,
        memory_search_model: BaseBackendLLM,
        extraction_model: BaseBackendLLM,
        enable_logging: bool = False,
    ):
        """
        Initialize the Memora instance.

        Args:
            vector_db (BaseVectorDB): Vector database.
            graph_db (BaseGraphDB): Graph database.
            memory_search_model (BaseBackendLLM): Model for memory search queries and Optional final filtering.
            extraction_model (BaseBackendLLM): Model for memory extraction operations.
            enable_logging (bool): Whether to enable console logging.

        Note:
            The graph database will be associated with the vector database.
        """

        self.memory_search_model = memory_search_model
        self.extraction_model = extraction_model
        self.vector_db = vector_db
        self.graph = graph_db

        # Associate the vector database with the graph database.
        self.graph.associated_vector_db = self.vector_db

        self.logger = logging.getLogger(__name__)
        if enable_logging:
            logging.basicConfig(level=logging.INFO)

    async def close(self) -> None:
        """Close and clean up resources used by Memora."""

        await self.vector_db.close()
        await self.graph.close()
        await self.memory_search_model.close()
        await self.extraction_model.close()
        self.logger.info("Memora resources cleaned.")

    async def generate_memory_search_queries(
        self,
        message: str,
        preceding_messages_for_context: List[Dict[str, str]] = [],
        current_datetime: datetime = datetime.now(),
    ) -> List[str]:
        """
        Generate memory search queries based on the given message and context.

        Args:
            message (str): The message to recall memories for.
            preceding_messages_for_context (List[Dict[str, str]]): Preceding messages for context.
            current_datetime (datetime): Current datetime.

        Returns:
            List[str]: List of generated memory search queries.
        """

        current_day_of_week = current_datetime.strftime("%A")
        response = await self.memory_search_model(
            messages=[
                {"role": "system", "content": MSG_MEMORY_SEARCH_PROMPT},
                {
                    "role": "user",
                    "content": MSG_MEMORY_SEARCH_TEMPLATE.format(
                        day_of_week=current_day_of_week,
                        current_datetime_str=current_datetime.isoformat(),
                        message_of_user=str(message),
                        preceding_messages=str(preceding_messages_for_context),
                    ),
                },
                {
                    "role": "assistant",
                    "content": "&& MEMORY_SEARCH &&",
                },  # For Guided Response.
            ]
        )

        arguments_pattern = re.compile(r"<<(.*?)>>", re.DOTALL)
        arguments_passed = arguments_pattern.findall(response)
        memory_search_queries = (
            [arg.strip() for arg in arguments_passed] if arguments_passed else []
        )

        self.logger.info(f"Generated memory search queries: {memory_search_queries}")

        return memory_search_queries

    async def filter_retrieved_memories_with_model(
        self,
        message: str,
        search_queries_used: List[str],
        retrieved_memories: List[models.Memory],
        current_datetime: datetime = datetime.now(),
    ) -> Set[str] | None:
        """
        Filter retrieved memories using the memory search model.

        Args:
            message (str): The message that triggered the search queries and retrieved memories.
            search_queries_used (List[str]): List of search queries used.
            retrieved_memories (List[Memory]): List of retrieved memories.
            current_datetime (datetime): Current datetime.

        Returns:
            Set[str] | None: Distinct Set of selected memory IDs (that passed the filter), or None if LLM was unable to analyze and select.
        """

        self.logger.info(f"Starting memory filtering for message: {message[:100]}...")
        self.logger.debug(f"Number of search queries used: {len(search_queries_used)}")
        self.logger.debug(
            f"Number of retrieved memories to filter: {len(retrieved_memories)}"
        )

        current_day_of_week = current_datetime.strftime("%A")
        self.logger.debug(
            f"Current day of week: {current_day_of_week}, datetime: {current_datetime.isoformat()}"
        )

        self.logger.info("Calling memory search model for filtering...")

        # Just use only needed information from retrieved memories.
        retrieved_memories: List[Dict[str, str]] = [
            memoryObj.id_memory_and_timestamp_dict() for memoryObj in retrieved_memories
        ]

        response = await self.memory_search_model(
            messages=[
                {
                    "role": "system",
                    "content": FILTER_RETRIEVED_MEMORIES_SYSTEM_PROMPT.format(
                        day_of_week=current_day_of_week,
                        current_datetime_str=current_datetime.isoformat(),
                        latest_room_message=str(message),
                        memory_search_queries="\n- ".join(search_queries_used),
                    ),
                },
                {"role": "user", "content": str(retrieved_memories)},
                {
                    "role": "assistant",
                    "content": "REASONS AND JUST memory_id enclosed in (<< >>):\n- Reason: ",
                },  # For Guided Response.
            ]
        )

        selected_id_pattern = re.compile(r"<<(.*?)>>", re.DOTALL)
        selected_memories_ids = selected_id_pattern.findall(response)

        if (
            not selected_memories_ids
        ):  # The LLM misbehaved not extracting any memory_ids or << NONE >>.
            self.logger.warning(
                "No memory IDs were extracted from the model response, due to LLM misbehavior."
            )
            return None

        # The LLM is undeterministic and can select the same memory_ids multiple times.
        filtered_ids = set(
            [
                str(selection).strip()
                for selection in selected_memories_ids
                if str(selection).strip().lower() != "none"
            ]
        )
        self.logger.info(
            f"Memory filtering complete. Selected {len(filtered_ids)} unique memories"
        )
        self.logger.debug(f"Selected memory IDs: {filtered_ids}")

        return filtered_ids

    async def search_memories_as_one(
        self,
        org_id: str,
        user_id: str,
        search_queries: List[str],
        filter_out_memory_ids_set: Set[str] = set(),
        agent_id: Optional[str] = None,
        search_across_agents: bool = True,
    ) -> List[models.Memory]:
        """
        Retrieve memories corresponding to the search queries for a message.

        Args:
            org_id (str): Organization ID.
            user_id (str): User ID.
            search_queries (List[str]): List of search queries.
            filter_out_memory_ids_set (Set[str]): Set of memory IDs to filter out.
            agent_id (Optional[str]): Agent ID.
            search_across_agents (bool): Whether to search memories across all agents.

        Returns:
            List[Memory]: List of retrieved memories.
        """

        self.logger.info(f"Searching memories for user {user_id} in org {org_id}")
        self.logger.debug(f"Search queries: {search_queries}")
        self.logger.debug(
            f"Agent context - agent_id: {agent_id}, memories_across_agents: {search_across_agents}"
        )

        batch_results = await self.vector_db.search_memories(
            queries=search_queries,
            memory_search_scope=MemorySearchScope.USER,
            org_id=org_id,
            user_id=user_id,
            agent_id=agent_id if not search_across_agents else None,
        )

        # Flatten and sort memories by score across the batch results
        sorted_memories = sorted(
            [
                memory_and_score
                for result in batch_results
                for memory_and_score in result
            ],
            key=lambda x: x[1],
            reverse=True,
        )

        # Extract the (org, user and memory ids), filtering out the ones to be excluded
        org_user_mem_ids = [
            {
                "memory_id": memory[0].memory_id,
                "user_id": memory[0].user_id,
                "org_id": memory[0].org_id,
            }
            for memory in sorted_memories
            if memory[0].memory_id not in filter_out_memory_ids_set
        ]

        if not org_user_mem_ids:
            self.logger.info(
                "No memories retrieved or left after filtering out `filter_out_memory_ids_set`"
            )
            return []

        return await self.graph.fetch_user_memories_resolved(org_user_mem_ids)

    async def search_memories_as_batch(
        self,
        org_id: str,
        search_queries: List[str],
        filter_out_memory_ids_set: Set[str] = set(),
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        memory_search_scope: MemorySearchScope = MemorySearchScope.USER,
        search_across_agents: bool = True,
    ) -> List[List[models.Memory]]:
        """
        Retrieve memories corresponding to a list of search queries.

        Args:
            org_id (str): Organization ID.
            search_queries (List[str]): List of search queries.
            filter_out_memory_ids_set (Set[str]): Set of memory IDs to filter out.
            user_id (Optional[str]): User ID.
            agent_id (Optional[str]): Agent ID.
            memory_search_scope (MemorySearchScope): Scope of memory search.
            search_across_agents (bool): Whether to search memories across all agents.

        Returns:
            List[List[Memory]]: Batch results of retrieved memories.
        """

        self.logger.info(f"Batch searching memories in org {org_id}")
        self.logger.debug(
            f"Search context - user_id: {user_id}, agent_id: {agent_id}, scope: {memory_search_scope}"
        )
        self.logger.debug(f"Number of search queries: {len(search_queries)}")

        batch_results = await self.vector_db.search_memories(
            queries=search_queries,
            memory_search_scope=memory_search_scope,
            org_id=org_id,
            user_id=user_id,
            agent_id=agent_id if not search_across_agents else None,
        )

        batch_org_user_mem_ids = [
            [
                {
                    "memory_id": memory[0].memory_id,
                    "user_id": memory[0].user_id,
                    "org_id": memory[0].org_id,
                }
                for memory in result
                if memory[0].memory_id not in filter_out_memory_ids_set
            ]
            for result in batch_results
        ]

        if not any(batch_org_user_mem_ids):
            self.logger.info(
                "No memories retrieved or left after filtering out `filter_out_memory_ids_set`"
            )
            return []

        return await self.graph.fetch_user_memories_resolved_batch(
            batch_org_user_mem_ids
        )

    async def save_or_update_interaction_and_memories(
        self,
        org_id: str,
        user_id: str,
        agent_id: str,
        interaction: List[Dict[str, str]],
        interaction_id: Optional[str] = None,
        current_datetime: datetime = datetime.now(),
        extract_agent_memories: bool = False,
        update_across_agents: bool = True,
        max_retries: int = 3,
    ) -> Tuple[str, datetime]:
        """
        Save a new interaction or update an existing one, and the extracted memories.

        FOR UPDATES -> Will Compare updated interaction with existing one:
            - If differences are found, truncates existing record from that point and
            replaces with updated version. Old memories from truncated message(s)
            remain but become standalone (no longer linked to truncated messages).
            - If no differences, appends new messages from the update.

        Args:
            org_id (str): Organization ID.
            user_id (str): User ID.
            agent_id (str): Agent ID.
            interaction (List[Dict[str, str]]): List of interaction messages.
            interaction_id (Optional[str]): Interaction ID for updates.
            current_datetime (datetime): Current datetime.
            extract_agent_memories (bool): Whether to extract agent memories.
            update_across_agents (bool): Whether to update memories across all agents.
            max_retries (int): Maximum number of retries.

        Returns:
            Tuple[str, datetime] containing:

                + interaction_id: Short UUID string
                + created_at or updated_at: Datetime object

        Raises:
            Exception: If saving the interaction and its memories fails after max retries.
        """
        operation = "Updating" if interaction_id else "Saving"
        self.logger.info(f"{operation} interaction for user {user_id} in org {org_id}")
        self.logger.debug(
            f"Interaction context - agent_id: {agent_id}, extract_agent_memories: {extract_agent_memories}"
        )
        self.logger.debug(f"Interaction size: {len(interaction)} messages")

        for retry in range(max_retries + 1):
            try:
                self.logger.debug(f"Attempt {retry + 1}/{max_retries + 1}")
                user, agent = await self._get_user_and_agent(org_id, user_id, agent_id)

                if interaction_id:
                    self.logger.debug(
                        f"Fetching previously extracted memories for interaction {interaction_id}"
                    )
                    previously_extracted_memories: List[Dict[str, str]] = [
                        memoryObj.memory_and_timestamp_dict()
                        for memoryObj in (
                            (
                                await self.graph.get_interaction(
                                    org_id,
                                    user_id,
                                    interaction_id,
                                    with_messages=False,
                                    with_memories=True,
                                )
                            ).memories
                            or []
                        )
                    ]

                    self.logger.debug(
                        f"Found {len(previously_extracted_memories)} previously extracted memories"
                    )

                    system_content = MEMORY_EXTRACTION_UPDATE_SYSTEM_PROMPT.format(
                        day_of_week=current_datetime.strftime("%A"),
                        current_datetime_str=current_datetime.isoformat(),
                        agent_label=agent.agent_label,
                        user_name=user.user_name,
                        extract_for_agent=(
                            f"and {agent.agent_label}" if extract_agent_memories else ""
                        ),
                        previous_memories=str(previously_extracted_memories),
                        schema=MemoryExtractionResponse.model_json_schema(),
                    )
                else:

                    system_content = MEMORY_EXTRACTION_SYSTEM_PROMPT.format(
                        day_of_week=current_datetime.strftime("%A"),
                        current_datetime_str=current_datetime.isoformat(),
                        agent_label=agent.agent_label,
                        user_name=user.user_name,
                        extract_for_agent=(
                            f"and {agent.agent_label}" if extract_agent_memories else ""
                        ),
                        schema=MemoryExtractionResponse.model_json_schema(),
                    )

                self.logger.debug("Preparing messages for memory extraction")
                messages = [{"role": "system", "content": system_content}]
                messages += [
                    {
                        "role": msg["role"],
                        "content": EXTRACTION_MSG_BLOCK_FORMAT.format(
                            message_id=i, content=msg["content"]
                        ),
                    }
                    for i, msg in enumerate(interaction)
                ]

                self.logger.info("Extracting memories from interaction")
                response: MemoryExtractionResponse = await self.extraction_model(
                    messages=messages, output_schema_model=MemoryExtractionResponse
                )

                candidate_memories, candidate_memories_msg_sources = (
                    self._process_extracted_memories(response, user, agent)
                )
                self.logger.debug(
                    f"Extracted {len(candidate_memories)} candidate memories"
                )

                if not candidate_memories:
                    self.logger.info("No useful information extracted for memories")
                    if interaction_id:
                        return await self.graph.update_interaction_and_memories(
                            org_id,
                            agent_id,
                            user_id,
                            interaction_id,
                            updated_memories_and_interaction=MemoriesAndInteraction(
                                interaction=interaction,
                                interaction_date=current_datetime,
                                memories=[],
                                contrary_memories=[],
                            ),
                        )
                    else:
                        return await self.graph.save_interaction_with_memories(
                            org_id,
                            agent_id,
                            user_id,
                            memories_and_interaction=MemoriesAndInteraction(
                                interaction=interaction,
                                interaction_date=current_datetime,
                                memories=[],
                                contrary_memories=[],
                            ),
                        )

                self.logger.info("Searching for existing related memories")
                existing_memories = await self.search_memories_as_one(
                    org_id=org_id,
                    user_id=user_id,
                    search_queries=candidate_memories,
                    agent_id=agent_id,
                    search_across_agents=update_across_agents,
                )

                if not existing_memories:
                    self.logger.info("No related existing memories found")
                    if interaction_id:
                        return await self.graph.update_interaction_and_memories(
                            org_id,
                            agent_id,
                            user_id,
                            interaction_id,
                            updated_memories_and_interaction=MemoriesAndInteraction(
                                interaction=interaction,
                                interaction_date=current_datetime,
                                memories=[
                                    MemoryToStore(
                                        memory=memory, source_msg_block_pos=source_msgs
                                    )
                                    for memory, source_msgs in zip(
                                        candidate_memories,
                                        candidate_memories_msg_sources,
                                    )
                                ],
                                contrary_memories=[],
                            ),
                        )
                    else:
                        return await self.graph.save_interaction_with_memories(
                            org_id,
                            agent_id,
                            user_id,
                            memories_and_interaction=MemoriesAndInteraction(
                                interaction=interaction,
                                interaction_date=current_datetime,
                                memories=[
                                    MemoryToStore(
                                        memory=memory, source_msg_block_pos=source_msgs
                                    )
                                    for memory, source_msgs in zip(
                                        candidate_memories,
                                        candidate_memories_msg_sources,
                                    )
                                ],
                                contrary_memories=[],
                            ),
                        )

                candidate_memories = [
                    {"memory": memory, "POS_ID": i}
                    for i, memory in enumerate(candidate_memories)
                ]
                messages = [
                    {
                        "role": "system",
                        "content": COMPARE_EXISTING_AND_NEW_MEMORIES_SYSTEM_PROMPT.format(
                            day_of_week=current_datetime.strftime("%A"),
                            current_datetime_str=current_datetime.isoformat(),
                            agent_placeholder=f"agent_{agent.agent_id}",
                            user_placeholder=f"user_{user.user_id}",
                            schema=MemoryComparisonResponse.model_json_schema(),
                        ),
                    },
                    {
                        "role": "user",
                        "content": COMPARE_EXISTING_AND_NEW_MEMORIES_INPUT_TEMPLATE.format(
                            existing_memories_string=str(existing_memories),
                            new_memories_string=str(candidate_memories),
                        ),
                    },
                ]

                response: MemoryComparisonResponse = await self.extraction_model(
                    messages=messages, output_schema_model=MemoryComparisonResponse
                )

                new_memories = []
                for memory in response.new_memories:
                    try:
                        new_memories.append(
                            (
                                memory.memory,
                                candidate_memories_msg_sources[
                                    memory.source_candidate_pos_id
                                ],
                            )
                        )
                    except Exception:
                        continue

                new_contrary_memories = []
                for memory in response.contrary_memories:
                    try:
                        new_contrary_memories.append(
                            (
                                memory.memory,
                                candidate_memories_msg_sources[
                                    memory.source_candidate_pos_id
                                ],
                                memory.contradicted_memory_id,
                            )
                        )
                    except Exception:
                        continue

                if interaction_id:
                    return await self.graph.update_interaction_and_memories(
                        org_id,
                        agent_id,
                        user_id,
                        interaction_id,
                        updated_memories_and_interaction=MemoriesAndInteraction(
                            interaction=interaction,
                            interaction_date=current_datetime,
                            memories=[
                                MemoryToStore(
                                    memory=memory_tuple[0],
                                    source_msg_block_pos=memory_tuple[1],
                                )
                                for memory_tuple in new_memories
                            ],
                            contrary_memories=[
                                ContraryMemoryToStore(
                                    memory=memory_tuple[0],
                                    source_msg_block_pos=memory_tuple[1],
                                    existing_contrary_memory_id=memory_tuple[2],
                                )
                                for memory_tuple in new_contrary_memories
                            ],
                        ),
                    )
                else:
                    return await self.graph.save_interaction_with_memories(
                        org_id,
                        agent_id,
                        user_id,
                        memories_and_interaction=MemoriesAndInteraction(
                            interaction=interaction,
                            interaction_date=current_datetime,
                            memories=[
                                MemoryToStore(
                                    memory=memory_tuple[0],
                                    source_msg_block_pos=memory_tuple[1],
                                )
                                for memory_tuple in new_memories
                            ],
                            contrary_memories=[
                                ContraryMemoryToStore(
                                    memory=memory_tuple[0],
                                    source_msg_block_pos=memory_tuple[1],
                                    existing_contrary_memory_id=memory_tuple[2],
                                )
                                for memory_tuple in new_contrary_memories
                            ],
                        ),
                    )

            except Exception:
                if retry == max_retries:
                    self.logger.error(
                        f"Failed to save/update interaction after {max_retries} retries",
                        exc_info=True,
                    )
                    raise
                else:
                    self.logger.warning(
                        f"Attempt {retry + 1} failed, retrying...", exc_info=True
                    )
                    continue

    async def _get_user_and_agent(
        self, org_id: str, user_id: str, agent_id: str
    ) -> Tuple[models.User, models.Agent]:
        """
        Fetch user and agent data.

        Args:
            org_id (str): Organization ID.
            user_id (str): User ID.
            agent_id (str): Agent ID.

        Returns:
            Tuple[User, Agent]: Tuple containing the user and agent data.
        """

        self.logger.debug(f"Fetching user {user_id} and agent {agent_id} data")
        user = await self.graph.get_user(org_id, user_id)
        agent = await self.graph.get_agent(org_id, agent_id)

        if not user or not agent:
            raise ValueError(
                f"User or Agent not found: user_id={user_id}, agent_id={agent_id}"
            )

        return user, agent

    def _process_extracted_memories(
        self, response: MemoryExtractionResponse, user: models.User, agent: models.Agent
    ) -> Tuple[List[str], List[List[int]]]:
        """
        Get every memory extracted, their source messages and insert placeholders for user and agent.

        Args:
            response (MemoryExtractionResponse): Response from the memory extraction model.
            user (User): User data.
            agent (Agent): Agent data.

        Returns:
            Tuple[List[str], List[List[int]]]: Extracted memories and their source messages.
        """

        candidate_memories: List[str] = []
        candidate_memories_msg_sources: List[List[int]] = []

        for memory_list in (
            response.memories_first_pass or [],
            response.memories_second_pass or [],
            response.memories_third_pass or [],
        ):
            for memory in memory_list:
                memory_text = memory.memory.replace(
                    "#user_#id#", f"user_{user.user_id}"
                ).replace("#agent_#id#", f"agent_{agent.agent_id}")
                candidate_memories.append(memory_text)
                candidate_memories_msg_sources.append(memory.msg_source_ids)

        return candidate_memories, candidate_memories_msg_sources

    async def recall_memories_for_message(
        self,
        org_id: str,
        user_id: str,
        latest_msg: str,
        agent_id: Optional[str] = None,
        preceding_msg_for_context: List[Dict[str, str]] = [],
        current_datetime: datetime = datetime.now(),
        filter_out_memory_ids_set: Set[str] = set(),
        search_memories_across_agents: bool = True,
        enable_final_model_based_memory_filter: bool = False,
    ) -> Tuple[List[models.Memory] | None, List[str] | None]:
        """
        Recall memories for the given message.

        Args:
            org_id (str): Organization ID.
            user_id (str): User ID.
            latest_msg (str): The latest message from user to find memories for.
            agent_id (Optional[str]): Agent ID.
            preceding_msg_for_context (List[Dict[str, str]]): Preceding messages for context.
            current_datetime (datetime): Current datetime.
            filter_out_memory_ids_set (Set[str]): Set of memory IDs to filter out.
            search_memories_across_agents (bool): Whether to search memories across all agents.
            enable_final_model_based_memory_filter (bool): 📝 Experimental feature; enables filtering of retrieved memories using a model. Note that a small model (~ 8B or lower) might not select some memories that are indirectly needed.

        Returns:
            Tuple[List[Memory] | None, List[str] | None]:

                + List[Memory]: Containing Memories to be recalled (None if no relevant memories found):

                    + org_id: Short UUID string identifying the organization
                    + agent_id: Short UUID string identifying the agent
                    + user_id: Short UUID string identifying the user
                    + interaction_id: Short UUID string identifying the interaction the memory was sourced from
                    + memory_id: Full UUID string identifying the memory
                    + memory: The resolved memory
                    + obtained_at: DateTime object of when the memory was obtained
                    + message_sources: List of messages in the interaction that triggered the memory

                + List[str]: Just the memory IDs (empty list [] if no memory was recalled).
        """

        self.logger.info(
            f"Getting memories for message from user {user_id} in org {org_id}"
        )
        self.logger.debug(
            f"Message context - agent_id: {agent_id}, context messages: {len(preceding_msg_for_context)}"
        )
        self.logger.debug(
            f"Model-based filtering enabled: {enable_final_model_based_memory_filter}"
        )

        self.logger.info("Generating memory search queries")
        search_queries = await self.generate_memory_search_queries(
            message=latest_msg,
            preceding_messages_for_context=preceding_msg_for_context,
            current_datetime=current_datetime,
        )
        self.logger.debug(f"Generated {len(search_queries)} search queries")

        if not search_queries:
            self.logger.warning("No search queries generated")
            search_queries = [latest_msg]  # Use the latest message as a fallback.

        self.logger.info("Searching memories based on generated queries")

        retrieved_memories = await self.search_memories_as_one(
            org_id=org_id,
            user_id=user_id,
            search_queries=search_queries,
            filter_out_memory_ids_set=filter_out_memory_ids_set,
            agent_id=agent_id,
            search_across_agents=search_memories_across_agents,
        )

        if not retrieved_memories:
            self.logger.info("No memories found for the message")
            return None, None

        self.logger.info(f"Retrieved {len(retrieved_memories)} memories")

        if not enable_final_model_based_memory_filter:
            return (
                retrieved_memories,  # memories.
                [memory.memory_id for memory in retrieved_memories],  # memory ids.
            )

        self.logger.info("Applying model-based memory filtering")
        filtered_memory_ids = await self.filter_retrieved_memories_with_model(
            latest_msg, search_queries, retrieved_memories, current_datetime
        )

        if (
            filtered_memory_ids is None
        ):  # The LLM was unable to filter just needed memories.
            self.logger.info("Model-based filtering failed")
            return (
                retrieved_memories,  # memories.
                [memory.memory_id for memory in retrieved_memories],  # memory ids.
            )

        if (
            len(filtered_memory_ids) == 0
        ):  # The LLM filtered out all memories (deemed none are needed to be recalled).
            self.logger.info("Model-based filtering returned no memories")
            return None, None

        memory_dict = {memory.memory_id: memory for memory in retrieved_memories}
        selected_memories = [
            memoryObj
            for memory_id in filtered_memory_ids
            if (memoryObj := memory_dict.get(memory_id)) is not None
        ]

        self.logger.info(
            f"Selected {len(selected_memories)} memories after model-based filtering"
        )
        return selected_memories, list(filtered_memory_ids)



import logging
import uuid
from datetime import datetime
from typing import List, Optional, Tuple

from qdrant_client import AsyncQdrantClient, models
from typing_extensions import override

import memora.schema.models as schema_models
from memora.vector_db.base import BaseVectorDB, MemorySearchScope


class QdrantDB(BaseVectorDB):

    def __init__(
        self,
        async_client: AsyncQdrantClient = None,
        collection_name: str = "memory_collection_v0_2",
        embed_models_cache_dir: str = "./cache",
        enable_logging: bool = False,
    ):
        """
        Initialize the QdrantDB class.

        Args:
            async_client (AsyncQdrantClient): A pre-initialized Async Qdrant client
            collection_name (str): Name of the Qdrant collection
            embed_models_cache_dir (str): Directory to cache the embedding models
            enable_logging (bool): Whether to enable console logging

        Example:
            ```python
            from qdrant_client import AsyncQdrantClient
            from memora.vector_db.qdrant import QdrantDB

            qdrant_db = QdrantDB(
                            async_client=AsyncQdrantClient(url="QDRANT_URL", api_key="QDRANT_API_KEY")
                        )
            ```
        """

        # Set Qdrant Client.
        self.async_client: AsyncQdrantClient = async_client

        # Set both dense and sparse embedding models to use for hybrid search.
        self.vector_embedding_model: str = (
            "sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2"
        )
        self.sparse_vector_embedding_model: str = "prithivida/Splade_PP_en_v1"

        self.async_client.set_model(
            self.vector_embedding_model, cache_dir=embed_models_cache_dir
        )
        self.async_client.set_sparse_model(
            self.sparse_vector_embedding_model, cache_dir=embed_models_cache_dir
        )

        # Set the collection name.
        self.collection_name = collection_name

        # Configure logging
        self.logger = logging.getLogger(__name__)
        if enable_logging:
            logging.basicConfig(level=logging.INFO)

    @override
    async def close(self) -> None:
        """Closes the qdrant database connection."""

        await self.async_client.close()
        self.logger.info("QdrantDB connection closed")

    # Setup methods
    @override
    async def setup(self, *args, **kwargs) -> None:
        """Setup the QdrantDB by creating the collection and payload indices."""

        await self._create_collection_if_not_exists()
        await self._create_payload_indices()
        self.logger.info("QdrantDB setup completed")

    async def _create_collection_if_not_exists(
        self, collection_name: Optional[str] = None
    ) -> None:
        """
        Create collection (if not exists) with vector and sparse vector configs for hybrid search.

        Args:
            collection_name (Optional[str]): Name for the Qdrant collection. Defaults to self.collection_name.
        """

        collection_name = collection_name or self.collection_name

        if not await self.async_client.collection_exists(collection_name):
            await self.async_client.create_collection(
                collection_name=collection_name,
                vectors_config=self.async_client.get_fastembed_vector_params(),
                sparse_vectors_config=self.async_client.get_fastembed_sparse_vector_params(),
                hnsw_config=models.HnswConfigDiff(
                    payload_m=16,
                    # Disable global index creation to utilize org_id and user_id
                    # payload indices for multi-tenancy.
                    m=0,
                ),
                quantization_config=models.ScalarQuantization(
                    scalar=models.ScalarQuantizationConfig(
                        type=models.ScalarType.INT8,
                        quantile=0.95,
                        # Store quantized vectors in RAM for faster access.
                        always_ram=True,
                    ),
                ),
            )
            self.logger.info(f"Created collection: {collection_name}")

    async def _create_payload_indices(self) -> None:
        """Create payload indices for multi-tenancy."""

        # Create org_id index
        await self.async_client.create_payload_index(
            collection_name=self.collection_name,
            field_name="org_id",
            field_schema=models.KeywordIndexParams(
                type="keyword",
                is_tenant=True,
            ),
        )

        # Create org_user_id index
        await self.async_client.create_payload_index(
            collection_name=self.collection_name,
            field_name="org_user_id",
            field_schema=models.KeywordIndexParams(
                type="keyword",
                is_tenant=True,
            ),
        )
        self.logger.info("Created payload indexes on vector DB.")

    # Embedding helper methods
    def _dense_embed_queries(self, queries: List[str]) -> List[List[float]]:
        """Embed queries using the dense vector embedding model."""
        return list(
            self.async_client.embedding_models[self.vector_embedding_model].embed(
                queries
            )
        )

    def _sparse_embed_queries(self, queries: List[str]) -> List[List[float]]:
        """Embed queries using the sparse vector embedding model."""
        return list(
            self.async_client.sparse_embedding_models[
                self.sparse_vector_embedding_model
            ].embed(queries)
        )

    # Core memory operations
    @override
    async def add_memories(
        self,
        org_id: str,
        user_id: str,
        agent_id: str,
        memory_ids: List[uuid.UUID],
        memories: List[str],
        obtained_at: str,
    ) -> None:
        """
        Add memories to collection with their org_id, user_id, agent_id, and obtained_at datetime as metadata.

        Args:
            org_id (str): Organization ID for the memories
            user_id (str): User ID for the memories
            agent_id (str): Agent ID for the memories
            memory_ids (List[uuid.UUID]): List of UUIDs for each memory
            memories (List[str]): List of memory strings to add
            obtained_at (str): ISO format datetime string when the memories were obtained
        """

        if not memories:
            raise ValueError("At least one memory and its memory id is required")

        if len(memories) != len(memory_ids):
            raise ValueError("Length of memories and memory_ids must match")

        metadata = [
            {
                "org_id": org_id,
                "org_user_id": f"{org_id}:{user_id}",
                "user_id": user_id,
                "agent_id": agent_id,
                "obtained_at": obtained_at,
            }
            for _ in memories
        ]

        await self.async_client.add(
            collection_name=self.collection_name,
            documents=memories,
            metadata=metadata,
            ids=[str(memory_id) for memory_id in memory_ids],
            # parallel=_  # Use all CPU cores
        )
        self.logger.info(
            f"Added {len(memories)} memories to collection: {self.collection_name}"
        )

    @override
    async def search_memory(
        self,
        query: str,
        memory_search_scope: MemorySearchScope,
        org_id: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> List[Tuple[schema_models.Memory, float]]:
        """
        Memory search with optional user/agent filtering.

        Args:
            query (str): Search query string
            memory_search_scope (MemorySearchScope): Memory search scope (organization or user)
            org_id (str): Organization ID for filtering
            user_id (Optional[str]): Optional user ID for filtering
            agent_id (Optional[str]): Optional agent ID for filtering

        Returns:
            List[Tuple[Memory, float]] containing tuple of search results and score:
                Memory:

                    + org_id: str
                    + agent_id: str
                    + user_id: str
                    + memory_id: str
                    + memory: str
                    + obtained_at: datetime

                float: Score of the memory
        """

        if not query:
            raise ValueError("A query is required")

        results = await self.search_memories(
            queries=[query],
            memory_search_scope=memory_search_scope,
            org_id=org_id,
            user_id=user_id,
            agent_id=agent_id,
        )
        return results[0] if results else []

    @override
    async def search_memories(
        self,
        queries: List[str],
        memory_search_scope: MemorySearchScope,
        org_id: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> List[List[Tuple[schema_models.Memory, float]]]:
        """
        Batch memory search with optional user/agent filtering.

        Args:
            queries (List[str]): List of search query strings
            memory_search_scope (MemorySearchScope): Memory search scope (organization or user)
            org_id (str): Organization ID for filtering
            user_id (Optional[str]): Optional user ID for filtering
            agent_id (Optional[str]): Optional agent ID for filtering

        Returns:
            List[List[Tuple[Memory, float]]] of search results for each query, with a tuple containing:
                Memory:

                    + org_id: str
                    + agent_id: str
                    + user_id: str
                    + memory_id: str
                    + memory: str
                    + obtained_at: datetime

                float: Score of the memory
        """

        if not queries:
            raise ValueError("At least one query is required")

        # Build filter conditions
        filter_conditions = []

        if (
            memory_search_scope == MemorySearchScope.ORGANIZATION
        ):  # Search memories across the organization.
            filter_conditions.append(
                models.FieldCondition(
                    key="org_id", match=models.MatchValue(value=org_id)
                )
            )
        elif (
            memory_search_scope == MemorySearchScope.USER
        ):  # Search memories for a specific user in an organization.

            if user_id is None:
                raise ValueError(
                    "user_id is required in addition to org_id for user-specific search"
                )

            filter_conditions.append(
                models.FieldCondition(
                    key="org_user_id",
                    match=models.MatchValue(value=f"{org_id}:{user_id}"),
                )
            )

        if (
            agent_id
        ):  # If agent id is provided, filter by agent also regardless of memory search scope.
            filter_conditions.append(
                models.FieldCondition(
                    key="agent_id", match=models.MatchValue(value=agent_id)
                )
            )

        # Embed queries
        dense_embeddings = self._dense_embed_queries(queries)
        sparse_embeddings = self._sparse_embed_queries(queries)

        search_results = await self.async_client.query_batch_points(
            collection_name=self.collection_name,
            requests=[
                models.QueryRequest(
                    prefetch=[
                        models.Prefetch(
                            query=models.SparseVector(
                                indices=sparse.indices, values=sparse.values
                            ),
                            using=self.async_client.get_sparse_vector_field_name(),
                            limit=12,
                        ),
                        models.Prefetch(
                            query=dense,
                            using=self.async_client.get_vector_field_name(),
                            score_threshold=0.4,
                            limit=12,
                        ),
                    ],
                    filter=(
                        models.Filter(must=filter_conditions)
                        if filter_conditions
                        else None
                    ),
                    with_payload=True,
                    query=models.FusionQuery(fusion=models.Fusion.RRF),
                    params=models.SearchParams(
                        quantization=models.QuantizationSearchParams(rescore=False)
                    ),
                )
                for sparse, dense in zip(sparse_embeddings, dense_embeddings)
            ],
        )

        search_results = [
            [
                (
                    schema_models.Memory(
                        org_id=point.payload["org_id"],
                        agent_id=point.payload["agent_id"],
                        user_id=point.payload["user_id"],
                        memory_id=point.id,
                        memory=point.payload["document"],
                        obtained_at=datetime.fromisoformat(
                            point.payload["obtained_at"]
                        ),
                    ),
                    point.score,
                )
                for point in query.points
                if point.score > 0.35  # Filter out low relevance memory.
            ]
            for query in search_results
        ]
        return search_results

    @override
    async def delete_memory(self, memory_id: str) -> None:
        """
        Delete a memory by its ID with optional org/user filtering.

        Args:
            memory_id (str): ID of the memory to delete
        """
        if memory_id:
            await self.async_client.delete(
                collection_name=self.collection_name,
                points_selector=models.PointIdsList(points=[memory_id]),
            )
            self.logger.info(f"Deleted memory with ID: {memory_id}")

    @override
    async def delete_memories(self, memory_ids: List[str]) -> None:
        """
        Delete multiple memories by their IDs.

        Args:
            memory_ids (List[str]): List of memory IDs to delete
        """

        if memory_ids:
            await self.async_client.delete(
                collection_name=self.collection_name,
                points_selector=models.PointIdsList(points=memory_ids),
            )
            self.logger.info(f"Deleted memories with IDs: {memory_ids}")

    @override
    async def delete_all_user_memories(self, org_id: str, user_id: str) -> None:
        """
        Delete all memories associated with a specific user.

        Args:
            org_id (str): Organization ID the user belongs to
            user_id (str): ID of the user whose memories should be deleted
        """

        await self.async_client.delete(
            collection_name=self.collection_name,
            points_selector=models.Filter(
                must=models.FieldCondition(
                    key="org_user_id",
                    match=models.MatchValue(value=f"{org_id}:{user_id}"),
                )
            ),
        )
        self.logger.info(
            f"Deleted all memories for user {user_id} in organization {org_id}"
        )

    @override
    async def delete_all_organization_memories(self, org_id: str) -> None:
        """
        Delete all memories associated with an organization.

        Args:
            org_id (str): ID of the organization whose memories should be deleted
        """

        filter_conditions = [
            models.FieldCondition(key="org_id", match=models.MatchValue(value=org_id))
        ]

        await self.async_client.delete(
            collection_name=self.collection_name,
            points_selector=models.Filter(must=filter_conditions),
        )
        self.logger.info(f"Deleted all memories for organization {org_id}")



from typing import List

import neo4j
import neo4j.exceptions
import shortuuid
from typing_extensions import override

from memora.schema import models

from ..base import BaseGraphDB


class Neo4jOrganization(BaseGraphDB):

    @override
    async def create_organization(self, org_name: str) -> models.Organization:
        """
        Creates a new organization in the Neo4j graph database.

        Args:
            org_name (str): The name of the organization to create.

        Returns:
            Organization object containing:

                + org_id: Short UUID string
                + org_name: Organization name
                + created_at: DateTime object of when the organization was created
        """

        if not isinstance(org_name, str) or not org_name:
            raise TypeError("`org_name` must be a string and have a value.")

        org_id = shortuuid.uuid()
        self.logger.info(f"Creating organization with ID {org_id}")

        async def create_org_tx(tx):
            result = await tx.run(
                """
                CREATE (o:Org {
                    org_id: $org_id,
                    org_name: $org_name,
                    created_at: datetime()
                })
                RETURN o{.org_id, .org_name, .created_at} as org
            """,
                org_id=org_id,
                org_name=org_name,
            )

            record = await result.single()
            return record["org"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:

            org_data = await session.execute_write(create_org_tx)

            if org_data is None:
                self.logger.info(f"Failed to create organization {org_id}")
                raise neo4j.exceptions.Neo4jError("Failed to create organization.")

            self.logger.info(f"Successfully created organization {org_id}")
            return models.Organization(
                org_id=org_data["org_id"],
                org_name=org_data["org_name"],
                created_at=(org_data["created_at"]).to_native(),
            )

    @override
    async def update_organization(
        self, org_id: str, new_org_name: str
    ) -> models.Organization:
        """
        Updates an existing organization in the Neo4j graph database.

        Args:
            org_id (str): The Short UUID of the organization to update.
            new_org_name (str): The new name for the organization.

        Returns:
            Organization object containing:

                + org_id: Short UUID string
                + org_name: Organization name
                + created_at: DateTime object of when the organization was created
        """

        if not all(
            param and isinstance(param, str) for param in (org_id, new_org_name)
        ):
            raise ValueError(
                "Both `org_id` and `new_org_name` must be a string and have a value."
            )

        self.logger.info(f"Updating organization {org_id}")

        async def update_org_tx(tx):
            result = await tx.run(
                """
                MATCH (o:Org {org_id: $org_id})
                SET o.org_name = $new_org_name
                RETURN o{.org_id, .org_name, .created_at} as org
            """,
                org_id=org_id,
                new_org_name=new_org_name,
            )

            record = await result.single()
            return record["org"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:

            org_data = await session.execute_write(update_org_tx)

            if org_data is None:
                self.logger.info(f"Organization {org_id} not found")
                raise neo4j.exceptions.Neo4jError(
                    "Organization (`org_id`) does not exist."
                )

            self.logger.info(f"Successfully updated organization {org_id}")
            return models.Organization(
                org_id=org_data["org_id"],
                org_name=org_data["org_name"],
                created_at=(org_data["created_at"]).to_native(),
            )

    @override
    async def delete_organization(self, org_id: str) -> None:
        """
        Deletes an organization from the Neo4j graph database.

        Warning:
            This operation will delete all nodes and relationships from this organization
            including users, agents, memories, interactions etc.

        Args:
            org_id (str): Short UUID string identifying the organization to delete.
        """

        if not isinstance(org_id, str) or not org_id:
            raise TypeError("`org_id` must be a string and have a value.")

        self.logger.info(f"Deleting organization {org_id} and all associated data")

        async def delete_org_tx(tx):
            # Delete all nodes and relationships associated with the org
            await tx.run(
                """
                CALL apoc.periodic.iterate("
                MATCH (o:Org {org_id: $org_id})
                CALL apoc.path.subgraphNodes(o, {}) YIELD node
                RETURN node",
                "DETACH DELETE node",
                {batchSize: 1000, parallel: true, params: {org_id: $org_id}})
            """,
                org_id=org_id,
            )

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(delete_org_tx)
            self.logger.info(f"Successfully deleted organization {org_id}")

    @override
    async def get_organization(self, org_id: str) -> models.Organization:
        """
        Gets a specific organization from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization to retrieve.

        Returns:
            Organization object containing:

                + org_id: Short UUID string
                + org_name: Organization name
                + created_at: DateTime object of when the organization was created
        """

        if not isinstance(org_id, str) or not org_id:
            raise TypeError("`org_id` must be a string and have a value.")

        async def get_org_tx(tx):
            result = await tx.run(
                """
                MATCH (o:Org {org_id: $org_id})
                RETURN o{.org_id, .org_name, .created_at} as org
            """,
                org_id=org_id,
            )
            record = await result.single()
            return record["org"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            org_data = await session.execute_read(get_org_tx)

            if org_data is None:
                self.logger.info(f"Organization {org_id} not found")
                raise neo4j.exceptions.Neo4jError(
                    "Organization (`org_id`) does not exist."
                )

            return models.Organization(
                org_id=org_data["org_id"],
                org_name=org_data["org_name"],
                created_at=(org_data["created_at"]).to_native(),
            )

    @override
    async def get_all_organizations(self) -> List[models.Organization]:
        """
        Gets all organizations from the graph database.

        Returns:
            List[Organization] each containing:

                + org_id: Short UUID string
                + org_name: Organization name
                + created_at: DateTime object of when the organization was created
        """

        self.logger.info("Getting all organizations")

        async def get_all_org_tx(tx):
            result = await tx.run(
                """
                MATCH (o:Org)
                RETURN o{.org_id, .org_name, .created_at} as org
            """
            )
            records = await result.value("org", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:

            all_org_data = await session.execute_read(get_all_org_tx)

            return [
                models.Organization(
                    org_id=org_data["org_id"],
                    org_name=org_data["org_name"],
                    created_at=(org_data["created_at"]).to_native(),
                )
                for org_data in all_org_data
            ]



from .filter_retrieved_memories import FILTER_RETRIEVED_MEMORIES_SYSTEM_PROMPT
from .memory_extraction import (
    COMPARE_EXISTING_AND_NEW_MEMORIES_INPUT_TEMPLATE,
    COMPARE_EXISTING_AND_NEW_MEMORIES_SYSTEM_PROMPT,
    EXTRACTION_MSG_BLOCK_FORMAT,
    MEMORY_EXTRACTION_SYSTEM_PROMPT,
    MEMORY_EXTRACTION_UPDATE_SYSTEM_PROMPT,
)
from .memory_search_from_msg import MSG_MEMORY_SEARCH_PROMPT, MSG_MEMORY_SEARCH_TEMPLATE

__all__ = [
    "FILTER_RETRIEVED_MEMORIES_SYSTEM_PROMPT",
    "MEMORY_EXTRACTION_UPDATE_SYSTEM_PROMPT",
    "MEMORY_EXTRACTION_SYSTEM_PROMPT",
    "EXTRACTION_MSG_BLOCK_FORMAT",
    "COMPARE_EXISTING_AND_NEW_MEMORIES_SYSTEM_PROMPT",
    "COMPARE_EXISTING_AND_NEW_MEMORIES_INPUT_TEMPLATE",
    "MSG_MEMORY_SEARCH_PROMPT",
    "MSG_MEMORY_SEARCH_TEMPLATE",
]



from typing import List, Optional

import neo4j
import neo4j.exceptions
import shortuuid
from typing_extensions import override

from memora.schema import models

from ..base import BaseGraphDB


class Neo4jAgent(BaseGraphDB):

    @override
    async def create_agent(
        self, org_id: str, agent_label: str, user_id: Optional[str] = None
    ) -> models.Agent:
        """
        Creates a new agent in the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            agent_label (str): Label/name for the agent.
            user_id (Optional[str]): Optional Short UUID of the user. This is used when the agent is created
                specifically for a user, indicating that both the organization and the
                user will have this agent.

        Returns:
            Agent containing:

                + org_id: Short UUID string
                + user_id: Optional Short UUID string
                + agent_id: Short UUID string
                + agent_label: Agent label/name
                + created_at: DateTime object of when the agent was created
        """

        if not all(param and isinstance(param, str) for param in (org_id, agent_label)):
            raise ValueError(
                "Both `org_id` and `agent_label` must be a string and have a value."
            )

        if user_id:
            if not isinstance(user_id, str):
                raise ValueError("`user_id` must be a string.")

        agent_id = shortuuid.uuid()
        self.logger.info(f"Creating new agent with ID {agent_id}")

        async def create_agent_tx(tx):
            if user_id:
                result = await tx.run(
                    """
                    MATCH (o:Org {org_id: $org_id}), (u:User {org_id: $org_id, user_id: $user_id})
                    CREATE (a:Agent {
                        org_id: $org_id,
                        user_id: $user_id,
                        agent_id: $agent_id,
                        agent_label: $agent_label,
                        created_at: datetime()
                    })
                    CREATE (o)-[:HAS_AGENT]->(a)
                    CREATE (u)-[:HAS_AGENT]->(a)
                    RETURN a{.org_id, .user_id, .agent_id, .agent_label, .created_at} as agent
                """,
                    org_id=org_id,
                    user_id=user_id,
                    agent_id=agent_id,
                    agent_label=agent_label,
                )
            else:
                result = await tx.run(
                    """
                    MATCH (o:Org {org_id: $org_id})
                    CREATE (a:Agent {
                        org_id: $org_id,
                        agent_id: $agent_id,
                        agent_label: $agent_label,
                        created_at: datetime()
                    })
                    CREATE (o)-[:HAS_AGENT]->(a)
                    RETURN a{.org_id, .agent_id, .agent_label, .created_at} as agent
                """,
                    org_id=org_id,
                    agent_id=agent_id,
                    agent_label=agent_label,
                )

            record = await result.single()
            return record["agent"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            agent_data = await session.execute_write(create_agent_tx)

            if agent_data is None:
                self.logger.info(f"Failed to create agent {agent_id}")
                raise neo4j.exceptions.Neo4jError("Failed to create agent.")

            self.logger.info(f"Successfully created agent {agent_id}")
            return models.Agent(
                org_id=agent_data["org_id"],
                agent_id=agent_data["agent_id"],
                user_id=agent_data.get("user_id"),
                agent_label=agent_data["agent_label"],
                created_at=(agent_data["created_at"]).to_native(),
            )

    @override
    async def update_agent(
        self, org_id: str, agent_id: str, new_agent_label: str
    ) -> models.Agent:
        """
        Updates an existing agent in the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            agent_id (str): Short UUID string identifying the agent to update.
            new_agent_label (str): New label/name for the agent.

        Returns:
            Agent containing:

                + org_id: Short UUID string
                + user_id: Optional Short UUID string
                + agent_id: Short UUID string
                + agent_label: Agent label/name
                + created_at: DateTime object of when the agent was created
        """

        if not all(
            param and isinstance(param, str)
            for param in (org_id, agent_id, new_agent_label)
        ):
            raise ValueError(
                "`org_id`, `agent_id` and `new_agent_name` must be strings and have a value."
            )

        self.logger.info(f"Updating agent {agent_id}")

        async def update_agent_tx(tx):
            result = await tx.run(
                """
                MATCH (a:Agent {org_id: $org_id, agent_id: $agent_id})
                SET a.agent_label = $new_agent_label
                RETURN a{.org_id, .user_id, .agent_id, .agent_label, .created_at} as agent
            """,
                org_id=org_id,
                agent_id=agent_id,
                new_agent_label=new_agent_label,
            )

            record = await result.single()
            return record["agent"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            agent_data = await session.execute_write(update_agent_tx)

            if agent_data is None:
                self.logger.info(
                    f"Failed to update agent {agent_id}: Agent does not exist"
                )
                raise neo4j.exceptions.Neo4jError(
                    "Agent (`org_id`, `agent_id`) does not exist."
                )

            self.logger.info(f"Successfully updated agent {agent_id}")
            return models.Agent(
                org_id=agent_data["org_id"],
                agent_id=agent_data["agent_id"],
                user_id=agent_data.get("user_id"),
                agent_label=agent_data["agent_label"],
                created_at=(agent_data["created_at"]).to_native(),
            )

    @override
    async def delete_agent(self, org_id: str, agent_id: str) -> None:
        """
        Deletes an agent from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            agent_id (str): Short UUID string identifying the agent to delete.
        """

        if not all(param and isinstance(param, str) for param in (org_id, agent_id)):
            raise ValueError(
                "`org_id` and `agent_id` must be strings and have a value."
            )

        self.logger.info(f"Deleting agent {agent_id}")

        async def delete_agent_tx(tx):
            # Using node key (org_id, agent_id) for faster lookup
            await tx.run(
                """
                MATCH (a:Agent {org_id: $org_id, agent_id: $agent_id})
                DETACH DELETE a
            """,
                org_id=org_id,
                agent_id=agent_id,
            )

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(delete_agent_tx)
            self.logger.info(f"Successfully deleted agent {agent_id}")

    @override
    async def get_agent(self, org_id: str, agent_id: str) -> models.Agent:
        """
        Gets a specific agent belonging to the specified organization from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            agent_id (str): Short UUID string identifying the agent to retrieve.

        Returns:
            Agent containing:

                + org_id: Short UUID string
                + user_id: Optional Short UUID string
                + agent_id: Short UUID string
                + agent_label: Agent label/name
                + created_at: DateTime object of when the agent was created
        """

        async def get_agent_tx(tx):
            result = await tx.run(
                """
                MATCH (a:Agent {org_id: $org_id, agent_id: $agent_id})
                RETURN a{.org_id, .user_id, .agent_id, .agent_label, .created_at} as agent
            """,
                org_id=org_id,
                agent_id=agent_id,
            )
            record = await result.single()
            return record["agent"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            agent_data = await session.execute_read(get_agent_tx)

            if agent_data is None:
                self.logger.info(
                    f"Failed to get agent {agent_id}: Agent does not exist"
                )
                raise neo4j.exceptions.Neo4jError(
                    "Agent (`org_id`, `agent_id`) does not exist."
                )

            return models.Agent(
                org_id=agent_data["org_id"],
                agent_id=agent_data["agent_id"],
                user_id=agent_data.get("user_id"),
                agent_label=agent_data["agent_label"],
                created_at=(agent_data["created_at"]).to_native(),
            )

    @override
    async def get_all_org_agents(self, org_id: str) -> List[models.Agent]:
        """
        Gets all agents belonging to the specified organization from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.

        Returns:
            A List[Agent], each containing:

                + org_id: Short UUID string
                + user_id: Optional Short UUID string
                + agent_id: Short UUID string
                + agent_label: Agent label/name
                + created_at: DateTime object of when the agent was created
        """

        if not isinstance(org_id, str) or not org_id:
            raise ValueError("`org_id` must be a string and have a value.")

        self.logger.info(f"Getting all agents for organization {org_id}")

        async def get_org_agents_tx(tx):
            result = await tx.run(
                """
                MATCH (o:Org {org_id: $org_id})-[:HAS_AGENT]->(a:Agent)
                RETURN a{.org_id, .user_id, .agent_id, .agent_label, .created_at} as agent
            """,
                org_id=org_id,
            )
            records = await result.value("agent", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            all_agents_data = await session.execute_read(get_org_agents_tx)

            return [
                models.Agent(
                    org_id=agent_data["org_id"],
                    agent_id=agent_data["agent_id"],
                    user_id=agent_data.get("user_id"),
                    agent_label=agent_data["agent_label"],
                    created_at=(agent_data["created_at"]).to_native(),
                )
                for agent_data in all_agents_data
            ]

    @override
    async def get_all_user_agents(
        self, org_id: str, user_id: str
    ) -> List[models.Agent]:
        """
        Gets all agents for a user within an organization from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            user_id (str): Short UUID string identifying the user.

        Returns:
            A List[Agent], each containing:

                + org_id: Short UUID string
                + user_id: Optional Short UUID string
                + agent_id: Short UUID string
                + agent_label: Agent label/name
                + created_at: DateTime object of when the agent was created
        """

        if not all(param and isinstance(param, str) for param in (org_id, user_id)):
            raise ValueError("`org_id` and `user_id` must be strings and have a value.")

        self.logger.info(
            f"Getting all agents for user {user_id} in organization {org_id}"
        )

        async def get_user_agents_tx(tx):
            result = await tx.run(
                """
                MATCH (u:User {org_id: $org_id, user_id: $user_id})-[:HAS_AGENT]->(a:Agent)
                RETURN a{.org_id, .user_id, .agent_id, .agent_label, .created_at} as agent
            """,
                org_id=org_id,
                user_id=user_id,
            )
            records = await result.value("agent", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            all_agents_data = await session.execute_read(get_user_agents_tx)

            return [
                models.Agent(
                    org_id=agent_data["org_id"],
                    agent_id=agent_data["agent_id"],
                    user_id=agent_data.get("user_id"),
                    agent_label=agent_data["agent_label"],
                    created_at=(agent_data["created_at"]).to_native(),
                )
                for agent_data in all_agents_data
            ]



import uuid
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Tuple

from memora.schema import models


class MemorySearchScope(Enum):
    ORGANIZATION = "organization"  # Search across all memories in the organization
    USER = "user"  # Search across all memories of a specific user in the organization


class BaseVectorDB(ABC):
    """
    Abstract base class defining a common interface for different Vector DB implementations.

    This class provides a standardized interface for vector database operations including
    adding, searching, and deleting memories.
    """

    @abstractmethod
    async def close(self) -> None:
        """Closes the database connection."""
        pass

    @abstractmethod
    async def setup(self, *args, **kwargs) -> None:
        """Setup the vector database by initializing collections, indices, etc."""
        pass

    @abstractmethod
    async def add_memories(
        self,
        org_id: str,
        user_id: str,
        agent_id: str,
        memory_ids: List[uuid.UUID],
        memories: List[str],
        obtained_at: str,
    ) -> None:
        """
        Add memories to collection with their org_id, user_id, agent_id, and obtained_at datetime as metadata.

        Args:
            org_id (str): Organization ID for the memories
            user_id (str): User ID for the memories
            agent_id (str): Agent ID for the memories
            memory_ids (List[uuid.UUID]): List of UUIDs for each memory
            memories (List[str]): List of memory strings to add
            obtained_at (str): ISO format datetime string when the memories were obtained
        """
        pass

    @abstractmethod
    async def search_memory(
        self,
        query: str,
        memory_search_scope: MemorySearchScope,
        org_id: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> List[Tuple[models.Memory, float]]:
        """
        Memory search with optional user/agent filtering.

        Args:
            query (str): Search query string
            memory_search_scope (MemorySearchScope): Memory search scope (organization or user)
            org_id (str): Organization ID for filtering
            user_id (Optional[str]): Optional user ID for filtering
            agent_id (Optional[str]): Optional agent ID for filtering

        Returns:
            List[Tuple[Memory, float]] containing tuple of search results and score:
                Memory:

                    + org_id: str
                    + agent_id: str
                    + user_id: str
                    + memory_id: str
                    + memory: str
                    + obtained_at: datetime

                float: Score of the memory
        """
        pass

    @abstractmethod
    async def search_memories(
        self,
        queries: List[str],
        memory_search_scope: MemorySearchScope,
        org_id: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> List[List[Tuple[models.Memory, float]]]:
        """
        Batch memory search with optional user/agent filtering.

        Args:
            queries (List[str]): List of search query strings
            memory_search_scope (MemorySearchScope): Memory search scope (organization or user)
            org_id (str): Organization ID for filtering
            user_id (Optional[str]): Optional user ID for filtering
            agent_id (Optional[str]): Optional agent ID for filtering

        Returns:
            List[List[Tuple[models.Memory, float]]] of search results for each query, with a tuple containing:
                Memory:

                    + org_id: str
                    + agent_id: str
                    + user_id: str
                    + memory_id: str
                    + memory: str
                    + obtained_at: datetime

                float: Score of the memory
        """
        pass

    @abstractmethod
    async def delete_memory(self, memory_id: str) -> None:
        """
        Delete a memory by its ID with optional org/user filtering.

        Args:
            memory_id (str): ID of the memory to delete
        """
        pass

    @abstractmethod
    async def delete_memories(self, memory_ids: List[str]) -> None:
        """
        Delete multiple memories by their IDs.

        Args:
            memory_ids (List[str]): List of memory IDs to delete
        """
        pass

    @abstractmethod
    async def delete_all_user_memories(self, org_id: str, user_id: str) -> None:
        """
        Delete all memories associated with a specific user.

        Args:
            org_id (str): Organization ID the user belongs to
            user_id (str): ID of the user whose memories should be deleted
        """
        pass

    @abstractmethod
    async def delete_all_organization_memories(self, org_id: str) -> None:
        """
        Delete all memories associated with an organization.

        Args:
            org_id (str): ID of the organization whose memories should be deleted
        """
        pass



from datetime import datetime

from pydantic import BaseModel, Field


class MemoryToStore(BaseModel):
    memory: str
    source_msg_block_pos: list[int] = Field(
        description="The position of the message block that resulted in this memory."
    )


class ContraryMemoryToStore(MemoryToStore):
    existing_contrary_memory_id: str = Field(
        description="The memory_id of the existing memory that was contradicted."
    )


class MemoriesAndInteraction(BaseModel):
    """
    Contains both the interaction, its date and the associated memories to store in memory stores.
    """

    interaction: list[dict[str, str]] = Field(
        default=[],
        description="The messages in the interaction [{'role': 'user', 'content': 'hello'}, ...]",
    )
    interaction_date: datetime = Field(
        default=datetime.now(),
        description="The date and time the interaction occurred.",
    )
    memories: list[MemoryToStore] = Field(
        default=[],
        description="The memories extracted from the interaction with their source messages position.",
    )
    contrary_memories: list[ContraryMemoryToStore] = Field(
        default=[],
        description="The memory extracted from the interaction with the above but also the memory id of the existing memory they contradicted.",
    )



from .agent import Neo4jAgent
from .interaction import Neo4jInteraction
from .interface import Neo4jGraphInterface
from .memory import Neo4jMemory
from .organization import Neo4jOrganization
from .user import Neo4jUser

__all__ = [
    "Neo4jGraphInterface",
    "Neo4jAgent",
    "Neo4jUser",
    "Neo4jInteraction",
    "Neo4jMemory",
    "Neo4jOrganization",
]



from typing import List

import neo4j
import neo4j.exceptions
import shortuuid
from typing_extensions import override

from memora.schema import models

from ..base import BaseGraphDB


class Neo4jUser(BaseGraphDB):

    @override
    async def create_user(self, org_id: str, user_name: str) -> models.User:
        """
        Creates a new user in the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            user_name (str): Name for the user.

        Returns:
            User containing:

                + org_id: Short UUID string
                + user_id: Short UUID string
                + user_name: User's name
                + created_at: DateTime object of when the user was created
        """

        if not all(param and isinstance(param, str) for param in (org_id, user_name)):
            raise ValueError(
                "Both `org_id` and `user_name` must be a string and have a value."
            )

        user_id = shortuuid.uuid()
        self.logger.info(f"Creating new user with ID {user_id}")

        async def create_user_tx(tx):
            result = await tx.run(
                """
                MATCH (o:Org {org_id: $org_id})
                CREATE (u:User {
                    org_id: $org_id,
                    user_id: $user_id,
                    user_name: $user_name,
                    created_at: datetime()
                })
                CREATE (u)-[:BELONGS_TO]->(o)
                CREATE (ic:InteractionCollection {
                    org_id: $org_id,
                    user_id: $user_id
                })
                CREATE (mc:MemoryCollection {
                    org_id: $org_id,
                    user_id: $user_id
                })
                CREATE (u)-[:INTERACTIONS_IN]->(ic)
                CREATE (u)-[:HAS_MEMORIES]->(mc)
                RETURN u{.org_id, .user_id, .user_name, .created_at} as user
            """,
                org_id=org_id,
                user_id=user_id,
                user_name=user_name,
            )
            record = await result.single()
            return record["user"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            user_data = await session.execute_write(create_user_tx)

            if user_data is None:
                self.logger.info(f"Failed to create user {user_id}")
                raise neo4j.exceptions.Neo4jError("Failed to create user.")

            return models.User(
                org_id=user_data["org_id"],
                user_id=user_data["user_id"],
                user_name=user_data["user_name"],
                created_at=(user_data["created_at"]).to_native(),
            )

    @override
    async def update_user(
        self, org_id: str, user_id: str, new_user_name: str
    ) -> models.User:
        """
        Updates an existing user in the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            user_id (str): Short UUID string identifying the user to update.
            new_user_name (str): The new name for the user.

        Returns:
            User containing:

                + org_id: Short UUID string
                + user_id: Short UUID string
                + user_name: User's name
                + created_at: DateTime object of when the user was created.
        """

        if not all(
            param and isinstance(param, str)
            for param in (org_id, user_id, new_user_name)
        ):
            raise ValueError(
                "`org_id`, `user_id` and `new_user_name` must be strings and have a value."
            )

        self.logger.info(f"Updating user {user_id}")

        async def update_user_tx(tx):
            result = await tx.run(
                """
                MATCH (u:User {org_id: $org_id, user_id: $user_id})
                SET u.user_name = $new_user_name
                RETURN u{.org_id, .user_id, .user_name, .created_at} as user
            """,
                org_id=org_id,
                user_id=user_id,
                new_user_name=new_user_name,
            )

            record = await result.single()
            return record["user"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            user_data = await session.execute_write(update_user_tx)

            if user_data is None:
                self.logger.info(
                    f"Failed to update user {user_id}: User does not exist"
                )
                raise neo4j.exceptions.Neo4jError(
                    "User (`org_id`, `user_id`) does not exist."
                )

            return models.User(
                org_id=user_data["org_id"],
                user_id=user_data["user_id"],
                user_name=user_data["user_name"],
                created_at=(user_data["created_at"]).to_native(),
            )

    @override
    async def delete_user(self, org_id: str, user_id: str) -> None:
        """
        Deletes a user from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            user_id (str): Short UUID string identifying the user to delete.
        """

        if not all(param and isinstance(param, str) for param in (org_id, user_id)):
            raise ValueError("`org_id` and `user_id` must be strings and have a value.")

        self.logger.info(f"Deleting user {user_id}")

        async def delete_user_tx(tx):
            await tx.run(
                """
                MATCH (u:User {org_id: $org_id, user_id: $user_id})
                OPTIONAL MATCH (u)-[:INTERACTIONS_IN]->(interactioncollection)
                OPTIONAL MATCH (interactioncollection)-[:HAD_INTERACTION]->(interaction)
                OPTIONAL MATCH (interaction)-[:FIRST_MESSAGE|IS_NEXT*]->(message)
                OPTIONAL MATCH (u)-[:HAS_MEMORIES]->(memcollection)
                OPTIONAL MATCH (memcollection)-[:INCLUDES]->(memory)
                OPTIONAL MATCH (interaction)-[:HAS_OCCURRENCE_ON]->(date)
                DETACH DELETE u, interactioncollection, interaction, message, memcollection, memory, date
            """,
                org_id=org_id,
                user_id=user_id,
            )

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(delete_user_tx)

    @override
    async def get_user(self, org_id: str, user_id: str) -> models.User:
        """
        Gets a specific user belonging to the specified organization from the Neo4j graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.
            user_id (str): Short UUID string identifying the user to retrieve.

        Returns:
            User containing:

                + org_id: Short UUID string
                + user_id: Short UUID string
                + user_name: User's name
                + created_at: DateTime object of when the user was created.
        """

        if not all(param and isinstance(param, str) for param in (org_id, user_id)):
            raise ValueError("`org_id` and `user_id` must be strings and have a value.")

        async def get_user_tx(tx):
            result = await tx.run(
                """
                MATCH (u:User {org_id: $org_id, user_id: $user_id})
                RETURN u{.org_id, .user_id, .user_name, .created_at} as user
            """,
                org_id=org_id,
                user_id=user_id,
            )
            record = await result.single()
            return record["user"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            user_data = await session.execute_read(get_user_tx)

            if user_data is None:
                self.logger.info(f"Failed to get user {user_id}: User does not exist")
                raise neo4j.exceptions.Neo4jError(
                    "User (`org_id`, `user_id`) does not exist."
                )

            return models.User(
                org_id=user_data["org_id"],
                user_id=user_data["user_id"],
                user_name=user_data["user_name"],
                created_at=(user_data["created_at"]).to_native(),
            )

    @override
    async def get_all_org_users(self, org_id: str) -> List[models.User]:
        """
        Gets all users belonging to the specified organization from the graph database.

        Args:
            org_id (str): Short UUID string identifying the organization.

        Returns:
            List[User], each containing:

                + org_id: Short UUID string
                + user_id: Short UUID string
                + user_name: User's name
                + created_at: DateTime object of when the user was created.
        """

        if not isinstance(org_id, str) or not org_id:
            raise ValueError("`org_id` must be a string and have a value.")

        self.logger.info(f"Getting all users for organization {org_id}")

        async def get_users_tx(tx):
            result = await tx.run(
                """
                MATCH (o:Org {org_id: $org_id})<-[:BELONGS_TO]-(u:User)
                RETURN u{.org_id, .user_id, .user_name, .created_at} as user
            """,
                org_id=org_id,
            )
            records = await result.value("user", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:

            all_users_data = await session.execute_read(get_users_tx)

            return [
                models.User(
                    org_id=user_data["org_id"],
                    user_id=user_data["user_id"],
                    user_name=user_data["user_name"],
                    created_at=(user_data["created_at"]).to_native(),
                )
                for user_data in all_users_data
            ]



from typing import Dict, List, Optional

import neo4j
from typing_extensions import override

from memora.schema import models

from ..base import BaseGraphDB


class Neo4jMemory(BaseGraphDB):

    @override
    async def fetch_user_memories_resolved(
        self, org_user_mem_ids: List[Dict[str, str]]
    ) -> List[models.Memory]:
        """
        Fetches memories from the Neo4j GraphDB by their IDs, resolves any contrary updates, and replaces user/agent placeholders with actual names.

        This method performs several operations:
          1. Retrieves memories using (org_id, user_id, memory_ids)
          2. If a memory has a CONTRARY_UPDATE relationship, uses the newer memory version
          3. Replaces user_id & agent_id placeholders (e.g 'user_abc123' or 'agent_xyz789') in memories with actual user names / agent labels

        Args:
            org_user_mem_ids (List[Dict[str, str]]): List of Dicts containing org, user, and memory ids of the memories to fetch and process

        Returns:
            List[Memory] containing memory details:

                + org_id: Short UUID string identifying the organization
                + agent_id: Short UUID string identifying the agent
                + user_id: Short UUID string identifying the user
                + interaction_id: Short UUID string identifying the interaction the memory was sourced from
                + memory_id: Full UUID string identifying the memory
                + memory: The resolved memory
                + obtained_at: DateTime object of when the memory was obtained
                + message_sources: List of messages in the interaction that triggered the memory

        Example:
            ```python
            >>> org_user_mem_ids = [{'memory_id': '443ac3a8-fe87-49a4-93d2-05d3eb58ddeb', 'org_id': 'gmDr4sUiWMNqbGAiV8ijbU', 'user_id': 'CcyKXxhi2skEcDpRzNZim7'}, ...]
            >>> memories = graphInstance.fetch_memories_resolved(org_user_mem_ids)
            >>> print([memoryObj.memory for memoryObj in memories])
            ["John asked for help with a wedding ring", "Sarah is allergic to peanuts"]
            ```

        Note:
            - Org, user, and memory IDs are typically retrieved from a vector database before being passed to this method.
            - A memory won't have a message source, if its interaction was updated with a conflicting conversation thread that lead to truncation of the former thread. See `graph.update_interaction_and_memories`
        """

        results = await self.fetch_user_memories_resolved_batch([org_user_mem_ids])
        return results[0]

    @override
    async def fetch_user_memories_resolved_batch(
        self, batch_org_user_mem_ids: List[List[Dict[str, str]]]
    ) -> List[List[models.Memory]]:
        """
        Fetches memories from the Neo4j GraphDB by their IDs, resolves any contrary updates, and replaces user/agent placeholders with actual names.

        This method performs several operations:
          1. Retrieves memories using (org_id, user_id, memory_ids)
          2. If a memory has a CONTRARY_UPDATE relationship, uses the newer memory version
          3. Replaces user_id & agent_id placeholders (e.g 'user_abc123' or 'agent_xyz789') in memories with actual user names / agent labels

        Args:
            batch_org_user_mem_ids (List[List[Dict[str, str]]]): List of lists containing Dicts with org, user, and memory ids of the memories to fetch and process

        Returns:
            List[List[Memory]] with memory details:

                + org_id: Short UUID string identifying the organization
                + agent_id: Short UUID string identifying the agent
                + user_id: Short UUID string identifying the user
                + interaction_id: Short UUID string identifying the interaction the memory was sourced from
                + memory_id: Full UUID string identifying the memory
                + memory: The resolved memory
                + obtained_at: DateTime object of when the memory was obtained
                + message_sources: List of messages in the interaction that triggered the memory

        Example:
            ```python
            >>> batch_org_user_mem_ids = [[{"memory_id": "413ac3a8-fe87-49a4-93d2-05d3eb58ddeb", "org_id": "gmDr4sUiWMNqbGAiV8ijbU", "user_id": "CcyKXxhi2skEcDpRzNZim7"}, ...], [{...}, ...]]
            >>> batch_memories = graphInstance.fetch_memories_resolved_batch(batch_org_user_mem_ids)
            >>> print([[memoryObj.memory for memoryObj in memories] for memories in batch_memories])
            [["John asked for help with a wedding ring", "Sarah is allergic to peanuts"], ["John is about to propose to Sarah"]]
            ```

        Note:
            - Batch org, user, and memory IDs are typically retrieved from a vector database before being passed to this method.
            - A memory won't have a message source, if its interaction was updated with a conflicting conversation thread that lead to truncation of the former thread. See `graph.update_interaction_and_memories`
        """

        async def fetch_resolved_batch_tx(tx):
            result = await tx.run(
                """
                UNWIND $batch_ids AS ids_dict_list

                CALL (ids_dict_list) {

                    UNWIND ids_dict_list AS ids_dict
                    MATCH (memory:Memory {org_id: ids_dict.org_id, user_id: ids_dict.user_id, memory_id: ids_dict.memory_id})
                                  
                    // Use the most up to date contrary update memory if it exists
                    OPTIONAL MATCH (memory)-[:CONTRARY_UPDATE*]->(contraryMemory:Memory) WHERE NOT (contraryMemory)-[:CONTRARY_UPDATE]->()
                    WITH coalesce(contraryMemory, memory) AS memoryToReturn

                    OPTIONAL MATCH (memoryToReturn)-[:MESSAGE_SOURCE]->(msgSource)
                    WITH memoryToReturn, collect(msgSource{.*}) as msgSources
                                  
                    MATCH (user:User {org_id: memoryToReturn.org_id, user_id: memoryToReturn.user_id})              
                    MATCH (agent:Agent {org_id: memoryToReturn.org_id, agent_id: memoryToReturn.agent_id})
                                  
                    // Case-insensitive 'user_' or 'agent_' followed by UUID and optional ('s) placeholders are replaced with actual names
                    RETURN collect(DISTINCT memoryToReturn{
                                                .org_id,
                                                .agent_id,
                                                .user_id,
                                                .interaction_id,
                                                .memory_id, 
                                                .obtained_at,
                                                memory: apoc.text.replace(
                                                    apoc.text.replace(memoryToReturn.memory, '(?i)user_[a-z0-9\\-]+(?:\\'s)?', user.user_name), 
                                                    '(?i)agent_[a-z0-9\\-]+(?:\\'s)?',  agent.agent_label
                                                ),
                                                message_sources: msgSources
                                            }) as resolved_memories
                }
                RETURN resolved_memories

            """,
                batch_ids=batch_org_user_mem_ids,
            )

            records = await result.value("resolved_memories", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:

            all_resolved_memories = await session.execute_read(fetch_resolved_batch_tx)

            return [
                [
                    models.Memory(
                        org_id=resolved_memory["org_id"],
                        agent_id=resolved_memory["agent_id"],
                        user_id=resolved_memory["user_id"],
                        interaction_id=resolved_memory["interaction_id"],
                        memory_id=resolved_memory["memory_id"],
                        memory=resolved_memory["memory"],
                        obtained_at=(resolved_memory["obtained_at"]).to_native(),
                        message_sources=[
                            models.MessageBlock(
                                role=msg_source["role"],
                                content=msg_source["content"],
                                msg_position=msg_source["msg_position"],
                            )
                            for msg_source in (resolved_memory["message_sources"] or [])
                        ],
                    )
                    for resolved_memory in resolved_memories
                ]
                for resolved_memories in all_resolved_memories
            ]

    @override
    async def get_user_memory(
        self, org_id: str, user_id: str, memory_id: str
    ) -> models.Memory:
        """
        Retrieves a specific memory.

        Args:
            org_id (str): Short UUID string identifying the organization
            user_id (str): Short UUID string identifying the user
            memory_id (str): UUID string identifying the memory

        Returns:
            Memory containing memory details:

                + org_id: Short UUID string identifying the organization
                + agent_id: Short UUID string identifying the agent
                + user_id: Short UUID string identifying the user
                + interaction_id: Short UUID string identifying the interaction the memory was sourced from
                + memory_id: Full UUID string identifying the memory
                + memory: The resolved memory
                + obtained_at: DateTime object of when the memory was obtained
                + message_sources: List of messages in the interaction that triggered the memory

        Note:
            - The memory won't have a message source, if its interaction was updated with a conflicting conversation thread that lead to truncation of the former thread. See `graph.update_interaction_and_memories`
        """

        if not all(
            param and isinstance(param, str) for param in (org_id, user_id, memory_id)
        ):
            raise ValueError(
                "`org_id`, `user_id` and `memory_id` must be strings and have a value."
            )

        self.logger.info(f"Getting memory {memory_id} for user {user_id}")

        async def get_memory_tx(tx):
            result = await tx.run(
                """
                MATCH (m:Memory {org_id: $org_id, user_id: $user_id, memory_id: $memory_id})

                MATCH (m)-[:MESSAGE_SOURCE]->(msgSource)
                WITH m, collect(msgSource{.*}) as msgSources

                MATCH (user:User {org_id: m.org_id, user_id: m.user_id})              
                MATCH (agent:Agent {org_id: m.org_id, agent_id: m.agent_id})
                RETURN m{
                        .org_id,
                        .agent_id,
                        .user_id,
                        .interaction_id,
                        .memory_id, 
                        .obtained_at,
                        memory: apoc.text.replace(
                            apoc.text.replace(m.memory, '(?i)user_[a-z0-9\\-]+(?:\\'s)?', user.user_name), 
                            '(?i)agent_[a-z0-9\\-]+(?:\\'s)?',  agent.agent_label
                        ),
                        message_sources: msgSources
                    } as memory
            """,
                org_id=org_id,
                user_id=user_id,
                memory_id=memory_id,
            )
            record = await result.single()
            return record["memory"] if record else None

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            memory = await session.execute_read(get_memory_tx)

            if not memory:
                self.logger.info(
                    f"Failed to get memory {memory_id}: Memory does not exist"
                )
                raise neo4j.exceptions.Neo4jError(
                    "Memory (`org_id`, `user_id`, `memory_id`) does not exist."
                )

            return models.Memory(
                org_id=memory["org_id"],
                agent_id=memory["agent_id"],
                user_id=memory["user_id"],
                interaction_id=memory["interaction_id"],
                memory_id=memory["memory_id"],
                memory=memory["memory"],
                obtained_at=(memory["obtained_at"]).to_native(),
                message_sources=[
                    models.MessageBlock(
                        role=msg_source["role"],
                        content=msg_source["content"],
                        msg_position=msg_source["msg_position"],
                    )
                    for msg_source in (memory["message_sources"] or [])
                ],
            )

    @override
    async def get_user_memory_history(
        self, org_id: str, user_id: str, memory_id: str
    ) -> List[models.Memory]:
        """
        Retrieves the history of a specific memory.

        Args:
            org_id (str): Short UUID string identifying the organization
            user_id (str): Short UUID string identifying the user
            memory_id (str): UUID string identifying the memory

        Returns:
            List[Memory] containing the history of memory details in descending order (starting with the current version, to the oldest version):

                + org_id: Short UUID string identifying the organization
                + agent_id: Short UUID string identifying the agent
                + user_id: Short UUID string identifying the user
                + interaction_id: Short UUID string identifying the interaction the memory was sourced from
                + memory_id: Full UUID string identifying the memory
                + memory: The resolved memory
                + obtained_at: DateTime object of when the memory was obtained
                + message_sources: List of messages in the interaction that triggered the memory

        Note:
            - A memory won't have a message source, if its interaction was updated with a conflicting conversation thread that lead to truncation of the former thread. See `graph.update_interaction_and_memories`
        """

        if not all(
            param and isinstance(param, str) for param in (org_id, user_id, memory_id)
        ):
            raise ValueError(
                "`org_id`, `user_id` and `memory_id` must be strings and have a value."
            )

        self.logger.info(f"Getting memory history for memory {memory_id}")

        async def get_memory_history_tx(tx):
            result = await tx.run(
                """
                MATCH path=(m:Memory {org_id: $org_id, user_id: $user_id, memory_id: $memory_id})<-[:CONTRARY_UPDATE*0..]-(olderMemory:Memory)
                WHERE NOT (olderMemory)<-[:CONTRARY_UPDATE]-()
                WITH nodes(path) AS memory_history
                UNWIND memory_history AS memory

                OPTIONAL MATCH (memory)-[:MESSAGE_SOURCE]->(msgSource)
                WITH memory, collect(msgSource{.*}) as msgSources

                MATCH (user:User {org_id: memory.org_id, user_id: memory.user_id})              
                MATCH (agent:Agent {org_id: memory.org_id, agent_id: memory.agent_id})
                RETURN memory{
                        .org_id,
                        .agent_id,
                        .user_id,
                        .interaction_id,
                        .memory_id, 
                        .obtained_at,
                        memory: apoc.text.replace(
                            apoc.text.replace(memory.memory, '(?i)user_[a-z0-9\\-]+(?:\\'s)?', user.user_name), 
                            '(?i)agent_[a-z0-9\\-]+(?:\\'s)?',  agent.agent_label
                        ),
                        message_sources: msgSources
                    } as memory
            """,
                org_id=org_id,
                user_id=user_id,
                memory_id=memory_id,
            )
            records = await result.value("memory", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            memory_history = await session.execute_read(get_memory_history_tx)
            return [
                models.Memory(
                    org_id=memory["org_id"],
                    agent_id=memory["agent_id"],
                    user_id=memory["user_id"],
                    interaction_id=memory["interaction_id"],
                    memory_id=memory["memory_id"],
                    memory=memory["memory"],
                    obtained_at=(memory["obtained_at"]).to_native(),
                    message_sources=[
                        models.MessageBlock(
                            role=msg_source["role"],
                            content=msg_source["content"],
                            msg_position=msg_source["msg_position"],
                        )
                        for msg_source in (memory["message_sources"] or [])
                    ],
                )
                for memory in memory_history
            ]

    @override
    async def get_all_user_memories(
        self,
        org_id: str,
        user_id: str,
        agent_id: Optional[str] = None,
        skip: int = 0,
        limit: int = 1000,
    ) -> List[models.Memory]:
        """
        Retrieves all memories associated with a specific user.

        Note:
            Memories are sorted in descending order by their obtained at datetime. (So most recent memories are first).

        Args:
            org_id (str): Short UUID string identifying the organization
            user_id (str): Short UUID string identifying the user
            agent_id (Optional[str]): Optional short UUID string identifying the agent. If provided, only memories obtained from
                interactions with this agent are returned.
                Otherwise, all memories associated with the user are returned.
            skip (int): Number of interactions to skip. (Useful for pagination)
            limit (int): Maximum number of interactions to retrieve. (Useful for pagination)

        Returns:
            List[Memory] containing memory details:

                + org_id: Short UUID string identifying the organization
                + agent_id: Short UUID string identifying the agent
                + user_id: Short UUID string identifying the user
                + interaction_id: Short UUID string identifying the interaction the memory was sourced from
                + memory_id: Full UUID string identifying the memory
                + memory: The resolved memory
                + obtained_at: DateTime object of when the memory was obtained
                + message_sources: List of messages in the interaction that triggered the memory

        Note:
            - A memory won't have a message source, if its interaction was updated with a conflicting conversation thread that lead to truncation of the former thread. See `graph.update_interaction_and_memories`
        """

        if not all(param and isinstance(param, str) for param in (org_id, user_id)):
            raise ValueError("`org_id` and `user_id` must be strings and have a value.")

        if agent_id:
            if not isinstance(agent_id, str):
                raise ValueError("`agent_id` must be a string.")

        self.logger.info(f" and agent {agent_id}" if agent_id else "")

        async def get_all_memories_tx(tx):
            query = """
                // Cleverly transverse through dates to get memories sorted, avoiding having to sort all user memory nodes.
                MATCH (d:Date {{org_id: $org_id, user_id: $user_id}})
                WITH d ORDER BY d.date DESC
                CALL (d) {{
                    MATCH (d)<-[:DATE_OBTAINED]-(memory)
                    {agent_filter}
                    RETURN memory ORDER BY memory.obtained_at DESC
                }}

                WITH memory AS m SKIP $skip LIMIT $limit

                OPTIONAL MATCH (m)-[:MESSAGE_SOURCE]->(msgSource)
                WITH m, collect(msgSource{{.*}}) as msgSources

                MATCH (user:User {{org_id: m.org_id, user_id: m.user_id}})
                MATCH (agent:Agent {{org_id: m.org_id, agent_id: m.agent_id}})

                RETURN m{{
                    .org_id,
                    .agent_id,
                    .user_id,
                    .interaction_id,
                    .memory_id, 
                    .obtained_at,
                    memory: apoc.text.replace(
                        apoc.text.replace(m.memory, '(?i)user_[a-z0-9\\-]+(?:\\'s)?', user.user_name), 
                        '(?i)agent_[a-z0-9\\-]+(?:\\'s)?',  agent.agent_label
                    ),
                    message_sources: msgSources
                }} as memory
            """
            agent_filter = "WHERE memory.agent_id = $agent_id" if agent_id else ""
            result = await tx.run(
                query.format(agent_filter=agent_filter),
                org_id=org_id,
                user_id=user_id,
                agent_id=agent_id,
                skip=skip,
                limit=limit,
            )

            records = await result.value("memory", [])
            return records

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.READ_ACCESS
        ) as session:
            memories = await session.execute_read(get_all_memories_tx)
            return [
                models.Memory(
                    org_id=memory["org_id"],
                    agent_id=memory["agent_id"],
                    user_id=memory["user_id"],
                    interaction_id=memory["interaction_id"],
                    memory_id=memory["memory_id"],
                    memory=memory["memory"],
                    obtained_at=(memory["obtained_at"]).to_native(),
                    message_sources=[
                        models.MessageBlock(
                            role=msg_source["role"],
                            content=msg_source["content"],
                            msg_position=msg_source["msg_position"],
                        )
                        for msg_source in (memory["message_sources"] or [])
                    ],
                )
                for memory in memories
            ]

    @override
    async def delete_user_memory(
        self,
        org_id: str,
        user_id: str,
        memory_id: str,
    ) -> None:
        """
        Deletes a specific memory.

        Args:
            org_id (str): Short UUID string identifying the organization
            user_id (str): Short UUID string identifying the user
            memory_id (str): UUID string identifying the memory to delete

        Note:
            If the graph database is associated with a vector database, the memory is also deleted there for data consistency.
        """

        if not all(
            param and isinstance(param, str) for param in (org_id, user_id, memory_id)
        ):
            raise ValueError(
                "`org_id`, `user_id` and `memory_id` must be strings and have a value."
            )

        self.logger.info(f"Deleting memory {memory_id}")

        async def delete_memory_tx(tx):
            await tx.run(
                """
                MATCH (m:Memory {org_id: $org_id, user_id: $user_id, memory_id: $memory_id})
                DETACH DELETE m
            """,
                org_id=org_id,
                user_id=user_id,
                memory_id=memory_id,
            )

            if (
                self.associated_vector_db and memory_id
            ):  # If the graph database is associated with a vector database
                # Delete memory from vector DB.
                await self.associated_vector_db.delete_memory(memory_id)

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(delete_memory_tx)
            self.logger.info(f"Successfully deleted memory {memory_id}")

    @override
    async def delete_all_user_memories(
        self,
        org_id: str,
        user_id: str,
    ) -> None:
        """
        Deletes all memories of a specific user.

        Args:
            org_id (str): Short UUID string identifying the organization
            user_id (str): Short UUID string identifying the user

        Note:
            If the graph database is associated with a vector database, the memories are also deleted there for data consistency.
        """

        if not all(param and isinstance(param, str) for param in (org_id, user_id)):
            raise ValueError("`org_id` and `user_id` must be strings and have a value.")

        self.logger.info(f"Deleting all memories for user {user_id}")

        async def delete_all_memories_tx(tx):
            await tx.run(
                """
                MATCH (u:User {org_id: $org_id, user_id: $user_id})-[:HAS_MEMORIES]->(mc:MemoryCollection)
                MATCH (mc)-[:INCLUDES]->(memory:Memory)
                DETACH DELETE memory
            """,
                org_id=org_id,
                user_id=user_id,
            )

            if (
                self.associated_vector_db
            ):  # If the graph database is associated with a vector database
                # Delete all memories from vector DB.
                await self.associated_vector_db.delete_all_user_memories(
                    org_id, user_id
                )

        async with self.driver.session(
            database=self.database, default_access_mode=neo4j.WRITE_ACCESS
        ) as session:
            await session.execute_write(delete_all_memories_tx)
            self.logger.info(f"Successfully deleted all memories for user {user_id}")



MEMORY_EXTRACTION_SYSTEM_PROMPT = """
The Current Date & Time is {day_of_week}, {current_datetime_str}.
Given an interaction between ({agent_label}) and ({user_name}).

# OBJECTIVE: 
- Extract details about {user_name} {extract_for_agent} from the interaction to be stored in memory used to personalize future interactions.
- Ignore insignificant details that you are very certain will be unhelpful in future personalized responses.
- Never fabricate any detail not present or strongly implied in the interaction.

# MEMORY GUIDELINES:  
- Keep each memory descriptive, self-contained, not exceed 25 words.  
- Use proper tense (past, present, continuous) as appropriate.  
- Always use #user_#id# instead of ({user_name}) and #agent_#id# instead of ({agent_label}) in the memories. 
- Output must be JSON format using the schema:
{schema}


>>>>>>> ENTIRE INTERACTION IS BELOW <<<<<<<
"""

EXTRACTION_MSG_BLOCK_FORMAT = """
# MESSAGE BLOCK ID: {message_id}
-------------
{content}
"""


MEMORY_EXTRACTION_UPDATE_SYSTEM_PROMPT = """ 
The Current Date & Time is {day_of_week}, {current_datetime_str}. 
Given an interaction between ({agent_label}) and ({user_name}). 

# OBJECTIVE:  
- Extract any new details about {user_name} {extract_for_agent} from the interaction that are not already present in previously extracted memories.
- Ignore insignificant details that you are very certain will be unhelpful in future personalized responses.
- Extract new information even if it contradicts previously extracted memories - contradictions are considered new information.
- Never fabricate any detail not present or strongly implied in the interaction. 

# MEMORY GUIDELINES:   
- Keep each memory descriptive, self-contained, not exceed 25 words.   
- Use proper tense (past, present, continuous) as appropriate.   
- Always use #user_#id# instead of ({user_name}) and #agent_#id# instead of ({agent_label}) in the memories. 
- Do not duplicate any information already present in the previously extracted memories. 
- If there is no useful new information or contrary update don't extract anything.
- Output must be JSON format using the schema: 
{schema}

# PREVIOUSLY ALREADY EXTRACTED MEMORIES:
{previous_memories}

>>>>>>> ENTIRE INTERACTION IS BELOW <<<<<<< 
"""

COMPARE_EXISTING_AND_NEW_MEMORIES_SYSTEM_PROMPT = """
The Current Date & Time is {day_of_week}, {current_datetime_str}.

You manage memories for {user_placeholder} / {agent_placeholder}.  
You are given existing stored memories and candidate new memories.  

# Objective: 
1. Identify New Memories:
   - Information gleaned from candidate memories that are not updates or redundant to any existing memory.

2. Identify Contradictory Memories:
   - Candidate memories that directly contradict existing stored memories.


# IMPORTANT GUIDELINES
- DO NOT GIVE ANY EXPLANATIONS.
- The Output JSON object must use the schema: {schema}

"""

COMPARE_EXISTING_AND_NEW_MEMORIES_INPUT_TEMPLATE = """
====
EXISTING MEMORIES 
====

{existing_memories_string}

====  
NEW CANDIDATE MEMORIES  
====  

{new_memories_string}

"""



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Type, Union

from pydantic import BaseModel


class BaseBackendLLM(ABC):
    """Abstract base class for LLMs used in the backend by Memora."""

    @abstractmethod
    async def close(self) -> None:
        """Closes the LLM connection."""
        pass

    @property
    @abstractmethod
    def get_model_kwargs(self) -> Dict[str, Any]:
        """
        Returns dictionary of model configuration parameters

        Example:
            return {
                "model": self.model, # model_name: gpt-4o
                "temperature": self.temperature, # 1
                "top_p": self.top_p, # 1
                "max_tokens": self.max_tokens, # 1024
                "stream": False,
            }
        """
        pass

    @abstractmethod
    async def __call__(
        self,
        messages: List[Dict[str, str]],
        output_schema_model: Type[BaseModel] | None = None,
    ) -> Union[str, BaseModel]:
        """
        Process messages and generate response (📌 Streaming is not supported, as full response is required at once)

        Args:
            messages (List[Dict[str, str]]): List of message dicts with role and content e.g [{"role": "user", "content": "Hello!"}, ...]
            output_schema_model (Type[BaseModel] | None): Optional Pydantic base model for structured output (📌 Ensure your model provider supports this for the chosen model)

        Returns:
            Union[str, BaseModel]: Generated text response as a string, or an instance of the output schema model if specified
        """
        pass



from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field


class Organization(BaseModel):
    org_id: str = Field(description="Short UUID string identifying the organization.")
    org_name: str = Field(description="Name of the organization.")
    created_at: datetime = Field(
        description="DateTime object of when the organization was created."
    )


class Agent(BaseModel):
    org_id: str = Field(description="Short UUID string identifying the organization.")
    agent_id: str = Field(description="Short UUID string identifying the agent.")
    user_id: Optional[str] = Field(
        default=None, description="Short UUID string identifying the user."
    )
    agent_label: str = Field(description="Label/name for the agent.")
    created_at: datetime = Field(
        description="DateTime object of when the agent was created."
    )


class User(BaseModel):
    org_id: str = Field(description="Short UUID string identifying the organization.")
    user_id: str = Field(description="Short UUID string identifying the user.")
    user_name: str = Field(description="Name of the user.")
    created_at: datetime = Field(
        description="DateTime object of when the user was created."
    )


class Interaction(BaseModel):
    org_id: str = Field(description="Short UUID string identifying the organization.")
    agent_id: str = Field(description="Short UUID string identifying the agent.")
    interaction_id: str = Field(
        description="Short UUID string identifying the interaction."
    )
    user_id: str = Field(description="Short UUID string identifying the user.")
    created_at: datetime = Field(
        description="DateTime object of when the interaction was created."
    )
    updated_at: datetime = Field(
        description="DateTime object of when the interaction was last updated."
    )
    messages: Optional[List[MessageBlock]] = Field(
        default=None, description="List of messages in the interaction."
    )
    memories: Optional[List[Memory]] = Field(
        default=None,
        description="List of memories gotten across all occurrences of this interaction.",
    )


class MessageBlock(BaseModel):
    role: Optional[str] = Field(
        default=None, description="e.g., user, assistant, tool."
    )
    content: Optional[str] = Field(
        default=None, description="The actual content of the message."
    )
    msg_position: int = Field(
        description="Position of this message in the interaction."
    )


class Memory(BaseModel):
    org_id: str = Field(description="Short UUID string identifying the organization.")
    agent_id: str = Field(description="Short UUID string identifying the agent.")
    user_id: str = Field(description="Short UUID string identifying the user.")
    interaction_id: Optional[str] = Field(
        default=None,
        description="Short UUID string identifying the interaction where this memory was sourced from.",
    )
    memory_id: str = Field(description="Full UUID string identifying the memory.")
    memory: str = Field(description="The memory.")
    obtained_at: datetime = Field(
        description="DateTime object of when the memory was obtained."
    )
    message_sources: Optional[List[MessageBlock]] = Field(
        default=None,
        description="List of messages in the interaction that triggered the memory. (Note: A memory won't have a message source, if its interaction was updated with a conflicting conversation thread that lead to truncation of the former thread. See `graph.update_interaction_and_memories`)",
    )

    def id_memory_and_timestamp_dict(self):
        return {
            "memory_id": self.memory_id,
            "memory": self.memory,
            "obtained_at": str(self.obtained_at),
        }

    def memory_and_timestamp_dict(self):
        return {"memory": self.memory, "obtained_at": str(self.obtained_at)}



MSG_MEMORY_SEARCH_PROMPT = """
You are a memory agent. Your task is to generate memory search queries based on the latest message to the room.

Input:
- Latest message to the room
- Previous conversation messages (if provided, and is useful for context)

Instructions:
1. Generate as many search queries needed to retrieve all relevant memories for the message (entities, their relationships, patterns, other info etc.)
2. Focus only on memory needs for the latest message
3. No explanations or responses - just search queries

Response Format:
&& MEMORY_SEARCH &&
    << Detailed memory search query >>
    << ... >>

Random Example (JUST EXAMPLE DO NOT USE ANY INFO HERE):
&& MEMORY_SEARCH &&
    << Who is Ava >>
    << Who is Sarah >>
    << What is Ava and Sarah Relationship >>
    << Ava vacation in San Francisco with Sarah >>
    << Ava and Sarah's visit to the Golden Gate Bridge >>
    << What is Sarah relationship with John and Who is he >>
    << Sarah planning a trip with John >>
"""

MSG_MEMORY_SEARCH_TEMPLATE = """
# Preceding Messages
---
{preceding_messages}
---

# Latest Message For Retrieval Decision (DateTime: {day_of_week}, {current_datetime_str})
---
{message_of_user}
---
"""



from typing import Any, Dict, List, Type, Union

from pydantic import BaseModel
from together import AsyncTogether
from typing_extensions import override

from .base import BaseBackendLLM


class TogetherBackendLLM(BaseBackendLLM):

    def __init__(
        self,
        api_key: str,
        model: str = "meta-llama/Llama-3.3-70B-Instruct-Turbo",
        temperature: float = 0.7,
        top_p: float = 1,
        max_tokens: int = 1024,
        max_retries: int = 3,
    ):
        """
        Initialize the TogetherBackendLLM class with specific parameters.

        Args:
            api_key (str): The API key to use for authentication
            model (str): The name of the Together model to use
            temperature (float): The temperature to use for sampling
            top_p (float): The top_p value to use for sampling
            max_tokens (int): The maximum number of tokens to generate
            max_retries (int): The maximum number of retries to make if a request fails

        Example:
            ```python
            from memora.llm_backends import TogetherBackendLLM

            together_backend_llm = TogetherBackendLLM(
                api_key="TOGETHER_API_KEY",
                model="meta-llama/Llama-3.3-70B-Instruct-Turbo"
            )
            ```
        """

        self.together_client = AsyncTogether(api_key=api_key, max_retries=max_retries)

        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.max_tokens = max_tokens

    @override
    async def close(self) -> None:
        """Closes the LLM connection."""

        self.together_client = None

    @override
    @property
    def get_model_kwargs(self) -> Dict[str, Any]:
        """Returns dictionary of model configuration parameters"""

        return {
            "model": self.model,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "max_tokens": self.max_tokens,
            "stream": False,
        }

    @override
    async def __call__(
        self,
        messages: List[Dict[str, str]],
        output_schema_model: Type[BaseModel] | None = None,
    ) -> Union[str, BaseModel]:
        """
        Process messages and generate response (📌 Streaming is not supported, as full response is required at once)

        Args:
            messages (List[Dict[str, str]]): List of message dicts with role and content e.g [{"role": "user", "content": "Hello!"}, ...]
            output_schema_model (Type[BaseModel] | None): Optional Pydantic base model for structured output (📌 Ensure the choosen model supports this)

        Returns:
            Union[str, BaseModel]: Generated text response as a string, or an instance of the output schema model if specified
        """

        if output_schema_model:
            response = await self.together_client.chat.completions.create(
                messages=messages,
                **self.get_model_kwargs,
                response_format={
                    "type": "json_object",
                    "schema": output_schema_model.model_json_schema(),
                },
            )
            content = response.choices[0].message.content
            return output_schema_model.model_validate_json(content)

        else:
            response = await self.together_client.chat.completions.create(
                messages=messages,
                **self.get_model_kwargs,
            )
            return response.choices[0].message.content



FILTER_RETRIEVED_MEMORIES_SYSTEM_PROMPT = """
The Current Date & Time is {day_of_week}, {current_datetime_str}.

This is the latest message sent to the room where an Agent and User are interacting:
Latest Message to Room: 
{latest_room_message}

These are the memory search queries based on the latest message:
Memory Search Queries: 
- {memory_search_queries}

You will receive the results of these memory search queries. Based on both the latest message and the results of the memory search queries, output the relevant memory_id (UUIDs) in the following format:

REASONS AND JUST memory_id enclosed in (<< >>):
- Reason: ... || << ... (just memory_id of a relevant memory here)>>
- Reason: ... || << ... >>

# If no relevant memory_id are found, output:
REASONS AND JUST memory_id enclosed in (<< >>):
- Reason: ... || << NONE >>
"""
